#
# Copyright (C) 2014 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import subprocess
import sys

pagerproc = None

def get_pager_env(name):
    if name not in os.environ:
        return None
    pager = os.environ[name]
    return pager

def get_pager():
    if not sys.stdout.isatty():
        return None

    pager = get_pager_env("GERRYMANDER_PAGER")
    if not pager:
        pager = get_pager_env("PAGER")
    if not pager:
        pager = "less"

    if pager == "cat":
        return None

    return pager

def stop_pager():
    global pagerproc
    if pagerproc is None:
        return

    sys.stdout.flush()
    sys.stderr.flush()
    os.close(1)
    pagerproc.stdin.close()
    pagerproc.wait()
    pagerproc = None

def start_pager():
    if not sys.stdout.isatty():
        return

    pager = get_pager()
    if not pager:
        return

    if "LESS" not in os.environ:
        os.environ["LESS"] = "FRSX"

    oldstdout = os.dup(1)
    global pagerproc
    pagerproc = subprocess.Popen([pager],
                                 stdin=subprocess.PIPE,
                                 stdout=oldstdout,
                                 close_fds=True)

    os.close(oldstdout)
    os.dup2(pagerproc.stdin.fileno(), 1)



#
# Copyright (C) 2014 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time

def format_date(then):
    if then is None:
        return ""
    try:
        now = time.time()
        delta = now - then
        return format_delta(delta)
    except (TypeError, ValueError):
        return ""

def format_delta(delta):
    days = delta / (60 * 60 * 24)
    hours = delta / (60 * 60)
    mins = delta / 60

    if days == 1:
        return "%d day" % days
    elif days > 1:
        return "%d days" % days
    elif hours == 1:
        return "%d hour" % hours
    elif hours > 1:
        return "%d hours" % hours
    elif mins == 1:
        return "%d min" % mins
    elif mins > 1:
        return "%d mins" % mins
    else:
        return "just now"


STYLES = {
    "reset": 0,
    "bold": 1,
    "underline": 4,
    "blinkslow": 5,
    "blinkfast": 6,
    "reverse": 7
}

COLORS = {
    "grey": 0,
    "red": 1,
    "green": 2,
    "yellow": 3,
    "blue": 4,
    "magenta": 5,
    "cyan": 6,
    "white": 7
}

FOREGROUND = 30
BACKGROUND = 40

ESCAPE = '\033[%dm'


def format_color(text, usecolor=True, fg=None, bg=None, styles=[]):
    if not usecolor:
        return text

    bits = []
    if fg is not None:
        if fg not in COLORS:
            raise Exception("Unknown color %s" % fg)
        bits.append(ESCAPE % (FOREGROUND + COLORS[fg]))
    if bg is not None:
        if bg not in COLORS:
            raise Exception("Unknown color %s" % bg)
        bits.append(ESCAPE % (BACKGROUND + COLORS[bg]))
    for style in styles:
        if style not in STYLES:
            raise Exception("Unknown style %s" % style)
        bits.append(ESCAPE % (STYLES[style]))
    bits.append(text)
    bits.append(ESCAPE % STYLES["reset"])
    return "".join(bits)


def format_title(text):
    width = len(text)
    underline = "-" * width
    return text + "\n" + underline + "\n"



#
# Copyright (C) 2014 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from gerrymander.client import ClientLive
from gerrymander.client import ClientCaching
from gerrymander.operations import OperationQuery
from gerrymander.operations import OperationWatch
from gerrymander.reports import ReportOutput
from gerrymander.reports import ReportPatchReviewStats
from gerrymander.reports import ReportPatchReviewRate
from gerrymander.reports import ReportOpenReviewStats
from gerrymander.reports import ReportChanges
from gerrymander.reports import ReportToDoListMine
from gerrymander.reports import ReportToDoListOthers
from gerrymander.reports import ReportToDoListAnyones
from gerrymander.reports import ReportToDoListNoones
from gerrymander.reports import ReportToDoListApprovable
from gerrymander.reports import ReportToDoListExpirable
from gerrymander.format import format_color
from gerrymander.model import ModelEventCommentAdd
from gerrymander.model import ModelEventPatchCreate
from gerrymander.model import ModelEventChangeMerge
from gerrymander.model import ModelEventChangeAbandon
from gerrymander.model import ModelEventChangeRestore
from gerrymander.model import ModelApproval
from gerrymander.pager import start_pager, stop_pager
from gerrymander.git import get_remote_info

import getpass
import os
import logging
import sys
import argparse
import textwrap

try:
    import configparser
except:
    import ConfigParser as configparser

class CommandConfig(object):

    def __init__(self, filename):
        self.filename = os.path.expanduser(filename)
        self.config = configparser.ConfigParser()
        self.config.read([self.filename])

        if self.has_option('server', 'remote'):
            self.set_gerrit_from_remote()

    def set_gerrit_from_remote(self):
        remote = self.get_option_string('server', 'remote')
        if not remote:
            return

        user, host, port = get_remote_info(remote)

        if host:
            self.config.set('server', 'hostname', host)

        if user:
            self.config.set('server', 'username', user)

        if port:
            self.config.set('server', 'port', port)

    def has_option(self, section, name):
        return self.config.has_option(section, name)

    def get_option_string(self, section, name, defvalue=None):
        if not self.config.has_option(section, name):
            return defvalue
        return self.config.get(section, name)

    def get_option_int(self, section, name, defvalue=None):
        if not self.config.has_option(section, name):
            return defvalue
        return int(self.config.get(section, name))

    def get_option_list(self, section, name, defvalue=None):
        if not self.config.has_option(section, name):
            return defvalue
        value = self.config.get(section, name)
        return list(map(lambda x: x.strip(), value.split(",")))

    def get_option_bool(self, section, name, defvalue=None):
        if not self.config.has_option(section, name):
            return defvalue
        value = self.config.get(section, name).strip()
        if value.lower() in ["true", "yes"]:
            return True
        return False

    def get_server_username(self):
        if not self.config.has_option("server", "username"):
            return None
        return self.config.get("server", "username")

    def get_server_hostname(self):
        if not self.config.has_option("server", "hostname"):
            return "review"
        return self.config.get("server", "hostname")

    def get_server_port(self):
        if not self.config.has_option("server", "port"):
            return 29418
        return self.config.get("server", "port")

    def get_server_keyfile(self):
        if not self.config.has_option("server", "keyfile"):
            return None
        return self.config.get("server", "keyfile")

    def get_cache_longlifetime(self):
        if not self.config.has_option("cache", "longlifetime"):
            return 86400
        return int(self.config.get("cache", "longlifetime"))

    def get_cache_shortlifetime(self):
        if not self.config.has_option("cache", "shortlifetime"):
            return 300
        return int(self.config.get("cache", "shortlifetime"))

    def get_cache_directory(self):
        if not self.config.has_option("cache", "directory"):
            return os.path.expanduser("~/.gerrymander.d/cache")
        return os.path.expanduser(self.config.get("cache", "directory"))

    def get_organization_groups(self):
        if not self.config.has_option("organization", "groups"):
            return []
        value = self.config.get("organization", "groups")
        return list(map(lambda x: x.strip(), value.split(",")))

    def get_organization_teams(self):
        if not self.config.has_option("organization", "teams"):
            return []
        value = self.config.get("organization", "teams")
        return list(map(lambda x: x.strip(), value.split(",")))

    def get_organization_bots(self):
        if not self.config.has_option("organization", "bots"):
            return []
        value = self.config.get("organization", "bots")
        return list(map(lambda x: x.strip(), value.split(",")))

    def get_group_projects(self, groupname):
        section = "group-" + groupname
        if not self.config.has_option(section, "projects"):
            return []
        value = self.config.get(section, "projects")
        return list(map(lambda x: x.strip(), value.split(",")))

    def get_group_team_members(self, groupname, teamname):
        section = "group-" + groupname
        key = "team-" + teamname
        if not self.config.has_option(section, key):
            return []
        value = self.config.get(section, key)
        return list(map(lambda x: x.strip(), value.split(",")))

    def get_command_aliases(self):
        if not self.config.has_option("commands", "aliases"):
            return []
        value = self.config.get("commands", "aliases")
        return list(map(lambda x: x.strip(), value.split(",")))

    def get_command_alias_basecmd(self, aliasname):
        section = "alias-" + aliasname
        return self.config.get(section, "basecmd")

    def get_command_alias_help(self, aliasname):
        section = "alias-" + aliasname
        return self.config.get(section, "help")


class Command(object):

    def __init__(self, name, help):
        super(Command, self).__init__()
        self.name = name
        self.help = help
        self.pager = True

    def add_option(self, parser, config, *args, **kwargs):
        if args[0][0:1] == "-":
            if args[0][0:2] == "--":
                name = args[0][2:]
            else:
                name = args[1][2:]
        else:
            name = args[0]

        section = "command-" + self.name
        if config.has_option(section, name):
            defvalue = kwargs["default"]
            if type(defvalue) == list:
                kwargs["default"] = config.get_option_list(section, name)
            elif type(defvalue) == int:
                kwargs["default"] = config.get_option_int(section, name)
            elif type(defvalue) == bool:
                kwargs["default"] = config.get_option_bool(section, name)
            else:
                kwargs["default"] = config.get_option_string(section, name)

        return parser.add_argument(*args, **kwargs)


    def add_options(self, parser, config):
        pass

    def get_client(self, config, options):
        return ClientLive(config.get_server_hostname(),
                          config.get_server_port(),
                          config.get_server_username(),
                          config.get_server_keyfile())

    def run(self, config, client, options):
        raise NotImplementedError("Subclass should override run method")

    def execute(self, config, options):
        if self.pager:
            start_pager()
        try:
            client = self.get_client(config, options)

            self.run(config, client, options)
        finally:
            if self.pager:
                stop_pager()


class CommandCaching(Command):

    def __init__(self, name, help):
        super(CommandCaching, self).__init__(name, help)
        self.longcache = False

    def set_long_cache(self, longcache):
        self.longcache = longcache

    def add_options(self, parser, config):
        super(CommandCaching, self).add_options(parser, config)
        self.add_option(parser, config,
                        "--no-cache", action="store_true",
                        help="Disable use of gerrit query cache")
        self.add_option(parser, config,
                        "--refresh", action="store_true",
                        help="Force refresh of the query cache")

    def get_client(self, config, options):
        if options.no_cache:
            return ClientLive(config.get_server_hostname(),
                              config.get_server_port(),
                              config.get_server_username(),
                              config.get_server_keyfile())
        else:
            if self.longcache:
                return ClientCaching(config.get_server_hostname(),
                                     config.get_server_port(),
                                     config.get_server_username(),
                                     config.get_server_keyfile(),
                                     os.path.join(config.get_cache_directory(), "long"),
                                     config.get_cache_longlifetime(),
                                     options.refresh)
            else:
                return ClientCaching(config.get_server_hostname(),
                                     config.get_server_port(),
                                     config.get_server_username(),
                                     config.get_server_keyfile(),
                                     os.path.join(config.get_cache_directory(), "short"),
                                     config.get_cache_shortlifetime(),
                                     options.refresh)


class CommandProject(Command):

    def __init__(self, name, help):
        super(CommandProject, self).__init__(name, help)


    def add_options(self, parser, config):
        super(CommandProject, self).add_options(parser, config)

        self.add_option(parser, config,
                        "-p", "--project", default=[],
                        action="append",
                        help="Gather information for project")

        self.add_option(parser, config,
                        "-g", "--group", default=[],
                        action="append",
                        help="Gather information for project group")

        self.add_option(parser, config,
                        "--all-groups", action="store_true",
                        help="Report on changes from all project groups")


    def get_projects(self, config, options, requireOne=False):
        count = 0
        if len(options.project) > 0:
            count = count + 1
        if len(options.group) > 0:
            count = count + 1
        if options.all_groups:
            count = count + 1

        if count > 1:
            raise Exception("--project, --group and --all-groups are mutually exclusive")
        if count == 0 and requireOne:
            raise Exception("One of --project, --group or --all-groups is required")

        if len(options.project) == 0:
            projects = []
            if options.all_groups:
                groups = config.get_organization_groups()
            else:
                groups = options.group

            for group in groups:
                projects.extend(config.get_group_projects(group))
            return projects
        else:
            return options.project


class CommandWatch(CommandProject):

    def __init__(self, name="watch", help="Watch incoming changes"):
        super(CommandWatch, self).__init__(name, help)

        self.pager = False

    def add_options(self, parser, config):
        super(CommandWatch, self).add_options(parser, config)

        self.add_option(parser, config,
                        "--color", default=False, action="store_true",
                        help="Use terminal color highlighting")
        self.add_option(parser, config,
                        "--all", default=False, action="store_true",
                        help="Don't filter list of comments to strip bots")

    @staticmethod
    def wrap_text(message, indent="", width=78):
        lines = textwrap.wrap(message, width)
        return "\n".join(map(lambda x: indent + x, lines))

    @staticmethod
    def format_comment(comment, user, usecolor):
        if comment == "":
            return
        print ("  %s (%s) wrote:" % (format_color(user.name,
                                                  usecolor,
                                                  styles=["bold"]),
                                     user.username))
        print ("")
        print (CommandComments.wrap_text(comment, "  "))

    @staticmethod
    def format_approvals(approvals):
        bits = []
        for approval in approvals:
            if approval.action == ModelApproval.ACTION_WORKFLOW and approval.value > 0:
                bits.append("+A")
            elif approval.action == ModelApproval.ACTION_REVIEWED:
                if approval.value > 0:
                    bits.append("R+" + str(approval.value))
                elif approval.value < 0:
                    bits.append("R" + str(approval.value))
                else:
                    bits.append("R=0")
            elif approval.action == ModelApproval.ACTION_VERIFIED:
                if approval.value > 0:
                    bits.append("V+" + str(approval.value))
                elif approval.value < 0:
                    bits.append("V" + str(approval.value))
                else:
                    bits.append("V=0")
        return ",".join(bits)

    @staticmethod
    def format_event(event, bots, projects, usecolor):
        if event.user is None or event.change is None:
            return

        if event.is_user_in_list(bots):
            return
        if len(projects) > 0 and event.change.project not in projects:
            return

        change = event.change
        print (format_color("Change %s (%s)" % (change.url, change.id),
                            usecolor,
                            fg="red",
                            styles=["bold"]))
        print ("")
        print ("  Project: %s" % change.project)
        print ("  Subject: %s" % change.subject)
        if type(event) == ModelEventChangeRestore:
            print ("   Action: %s" % format_color("change restored", styles=["bold"]))
        elif type(event) == ModelEventChangeAbandon:
            print ("   Action: %s" % format_color("change abandoned", styles=["bold"]))
        elif type(event) == ModelEventChangeMerge:
            print ("   Action: %s" % format_color("change merged", styles=["bold"]))
        elif type(event) == ModelEventCommentAdd:
            print ("   Action: %s" % format_color("comment added", styles=["bold"]))
            if len(event.approvals) > 0:
                print ("    Votes: %s" %
                       format_color(CommandWatch.format_approvals(event.approvals),
                                    usecolor,
                                    fg="blue",
                                    styles=["bold"]))

        elif type(event) == ModelEventPatchCreate:
            print ("   Action: %s" % format_color("change restored", styles=["bold"]))

        if type(event) == ModelEventCommentAdd:
            print ("")
            CommandWatch.format_comment(event.comment, event.user, usecolor)
        print ("")
        print ("")


    def run(self, config, client, options):
        watch = OperationWatch(client)

        if options.all:
            bots = []
        else:
            bots = config.get_organization_bots()

        projects = self.get_projects(config, options)

        def cb(event):
            self.format_event(event,
                              bots,
                              projects,
                              options.color)

        return watch.run(cb)


class CommandReport(Command):

    def __init__(self, name, help):
        super(CommandReport, self).__init__(name, help)

    def add_options(self, parser, config):
        super(CommandReport, self).add_options(parser, config)

        self.add_option(parser, config,
                        "-m", "--mode", default=ReportOutput.DISPLAY_MODE_TEXT,
                        help="Display output in 'text', 'json', 'xml', 'csv'")
        self.add_option(parser, config,
                        "--color", default=False, action="store_true",
                        help="Use terminal color highlighting")

    def get_report(self, config, client, options):
        raise NotImplementedError("subclass must override get_query")

    def run(self, config, client, options):
        report = self.get_report(config, client, options)

        report.display(options.mode)


class CommandReportTable(CommandReport):

    def __init__(self, name, help):
        super(CommandReportTable, self).__init__(name, help)

    def add_options(self, parser, config):
        super(CommandReportTable, self).add_options(parser, config)

        self.add_option(parser, config,
                        "-l", "--limit", default=None,
                        help="Limit to N results")

        self.sort_option = self.add_option(parser, config,
                                           "--sort", default=None,
                                           help="Set the sort field")
        self.add_option(parser, config,
                        "--field", default=[],
                        action="append",
                        help="Display the named field")

    def run(self, config, client, options):
        report = self.get_report(config, client, options)

        if options.limit is not None:
            report.set_data_limit(int(options.limit))

        if options.sort is not None:
            reverse = False
            offset = options.sort.find(":")
            if offset != -1:
                if options.sort[offset + 1:] == "rev":
                    reverse = True
                options.sort = options.sort[0:offset]
            report.set_sort_column(options.sort, reverse)

        if len(options.field) > 0:
            for col in report.get_columns():
                col.visible = False
            for key in options.field:
                name = key
                offset = name.find(":")
                truncate = None
                if offset != -1:
                    truncate = int(name[offset+1:])
                    name = name[0:offset]
                col = report.get_column(name)
                if col is None:
                    raise Exception("Unknown field '%s'" % name)
                if truncate is not None:
                    col.truncate = truncate
                col.visible = True

        report.display(options.mode)


class CommandPatchReviewStats(CommandProject, CommandCaching, CommandReportTable):

    def __init__(self, name="patchreviewstats", help="Statistics on patch review approvals"):
        super(CommandPatchReviewStats, self).__init__(name, help)
        self.teams = {}
        self.set_long_cache(True)

    def add_options(self, parser, config):
        super(CommandPatchReviewStats, self).add_options(parser, config)

        self.add_option(parser, config,
                        "--days", default=30,
                        help="Set number of days history to consult")

    def get_report(self, config, client, options):
        if options.all_groups:
            groups = config.get_organization_groups()
        else:
            groups = options.group

        teams = {}
        for team in config.get_organization_teams():
            teams[team] = []
            for group in groups:
                users = config.get_group_team_members(group, team)
                teams[team].extend(users)

        return ReportPatchReviewStats(client,
                                      self.get_projects(config, options, True),
                                      int(options.days),
                                      teams,
                                      usecolor=options.color)

    def run(self, config, client, options):
        return super(CommandPatchReviewStats, self).run(config, client, options)


class CommandPatchReviewRate(CommandProject, CommandCaching, CommandReportTable):

    def __init__(self, name="patchreviewrate", help="Daily review rate averaged per week"):
        super(CommandPatchReviewRate, self).__init__(name, help)
        self.teams = {}
        self.set_long_cache(True)

    def get_report(self, config, client, options):
        if options.all_groups:
            groups = config.get_organization_groups()
        else:
            groups = options.group

        teams = {}
        for team in config.get_organization_teams():
            teams[team] = []
            for group in groups:
                users = config.get_group_team_members(group, team)
                teams[team].extend(users)

        return ReportPatchReviewRate(client,
                                     self.get_projects(config, options, True),
                                     teams,
                                     usecolor=options.color)

    def run(self, config, client, options):
        return super(CommandPatchReviewRate, self).run(config, client, options)


class CommandOpenReviewStats(CommandProject, CommandCaching, CommandReportTable):

    def __init__(self, name="openreviewstats", help="Statistics on open patch reviews"):
        super(CommandOpenReviewStats, self).__init__(name, help)
        self.teams = {}
        self.set_long_cache(True)

    def add_options(self, parser, config):
        super(CommandOpenReviewStats, self).add_options(parser, config)

        self.add_option(parser, config,
                        "--branch", default="master",
                        help="Set branch name to query")
        self.add_option(parser, config,
                        "--days", default=7,
                        help="Show count waiting more than N days")
        self.add_option(parser, config,
                        "--topic", default="",
                        help="Set topic name to query")

    def get_report(self, config, client, options):
        return ReportOpenReviewStats(client,
                                     self.get_projects(config, options, True),
                                     options.branch,
                                     options.topic,
                                     int(options.days),
                                     usecolor=options.color)

    def run(self, config, client, options):
        if options.limit is None:
            options.limit = 5

        return super(CommandOpenReviewStats, self).run(config, client, options)


class CommandChanges(CommandProject, CommandCaching, CommandReportTable):

    def __init__(self, name="changes", help="Query project changes"):
        super(CommandChanges, self).__init__(name, help)

    def add_options(self, parser, config):
        super(CommandChanges, self).add_options(parser, config)

        self.add_option(parser, config,
                        "--status", action="append", default=[],
                        help="Filter based on status")
        self.add_option(parser, config,
                        "--reviewer", action="append", default=[],
                        help="Filter based on reviewer")
        self.add_option(parser, config,
                        "--branch", action="append", default=[],
                        help="Filter based on branch")
        self.add_option(parser, config,
                        "--topic", action="append", default=[],
                        help="Filter based on topic")
        self.add_option(parser, config,
                        "--message", action="append", default=[],
                        help="Filter based on message")
        self.add_option(parser, config,
                        "--owner", action="append", default=[],
                        help="Filter based on owner")
        self.add_option(parser, config,
                        "--approval", action="append", default=[],
                        help="Filter based on approval")
        self.add_option(parser, config,
                        "--rawquery", default=None,
                        help="Raw query string to pass through to gerrit")
        self.add_option(parser, config,
                        "--deps", action="store_true", default=False,
                        help="Sort output by change dependencies")
        self.add_option(parser, config,
                        "file", default=[], nargs="*",
                        help="File name matches")
        self.sort_option.choices = [c.key for c in ReportChanges.COLUMNS]

    def get_report(self, config, client, options):
        return ReportChanges(client,
                             self.get_projects(config, options),
                             status=options.status,
                             reviewers=options.reviewer,
                             branches=options.branch,
                             topics=options.topic,
                             messages=options.message,
                             owners=options.owner,
                             approvals=options.approval,
                             rawquery=options.rawquery,
                             files=options.file,
                             usecolor=options.color,
                             deps=options.deps)


class CommandToDo(CommandProject, CommandCaching, CommandReportTable):
    def __init__(self, name, help):
        super(CommandToDo, self).__init__(name, help)

    def add_options(self, parser, config):
        super(CommandToDo, self).add_options(parser, config)

        self.add_option(parser, config,
                        "--branch", action="append", default=[],
                        help="Filter based on branch")
        self.add_option(parser, config,
                        "--topic", action="append", default=[],
                        help="Filter based on topic")
        self.add_option(parser, config,
                        "--deps", action="store_true", default=False,
                        help="Sort output by change dependencies")
        self.add_option(parser, config,
                        "file", default=[], nargs="*",
                        help="File name matches")

class CommandToDoMine(CommandToDo):

    def __init__(self, name="todo-mine", help="List of changes I've looked at before"):
        super(CommandToDoMine, self).__init__(name, help)

    def get_report(self, config, client, options):
        username = config.get_server_username()
        if username is None:
            username = getpass.getuser()

        return ReportToDoListMine(client,
                                  username=username,
                                  projects=self.get_projects(config, options),
                                  branches=options.branch,
                                  files=options.file,
                                  topics=options.topic,
                                  usecolor=options.color,
                                  deps=options.deps)


class CommandToDoOthers(CommandToDo):

    def __init__(self, name="todo-others", help="List of changes I've not looked at before"):
        super(CommandToDoOthers, self).__init__(name, help)

    def get_report(self, config, client, options):
        username = config.get_server_username()
        if username is None:
            username = getpass.getuser()

        return ReportToDoListOthers(client,
                                    username=username,
                                    projects=self.get_projects(config, options),
                                    branches=options.branch,
                                    files=options.file,
                                    topics=options.topic,
                                    usecolor=options.color,
                                    deps=options.deps)


class CommandToDoAnyones(CommandToDo):

    def __init__(self, name="todo-anyones", help="List of changes anyone has looked at"):
        super(CommandToDoAnyones, self).__init__(name, help)

    def get_report(self, config, client, options):
        username = config.get_server_username()
        if username is None:
            username = getpass.getuser()

        return ReportToDoListAnyones(client,
                                     username=username,
                                     bots=config.get_organization_bots(),
                                     projects=self.get_projects(config, options),
                                     branches=options.branch,
                                     files=options.file,
                                     topics=options.topic,
                                     usecolor=options.color,
                                     deps=options.deps)


class CommandToDoNoones(CommandToDo):

    def __init__(self, name="todo-noones", help="List of changes no one has looked at yet"):
        super(CommandToDoNoones, self).__init__(name, help)

    def get_report(self, config, client, options):
        return ReportToDoListNoones(client,
                                    bots=config.get_organization_bots(),
                                    projects=self.get_projects(config, options),
                                    branches=options.branch,
                                    files=options.file,
                                    topics=options.topic,
                                    usecolor=options.color,
                                    deps=options.deps)


class CommandToDoApprovable(CommandToDo):

    def __init__(self, name="todo-approvable", help="List of changes that I can approve"):
        super(CommandToDoApprovable, self).__init__(name, help)

    def add_options(self, parser, config):
        super(CommandToDoApprovable, self).add_options(parser, config)

        self.add_option(parser, config,
                        "--strict", action="store_true", default=False,
                        help="Exclude changes with any negative code reviews")

    def get_report(self, config, client, options):
        username = config.get_server_username()
        if username is None:
            username = getpass.getuser()

        return ReportToDoListApprovable(client,
                                        username=username,
                                        strict=options.strict,
                                        projects=self.get_projects(config, options),
                                        branches=options.branch,
                                        files=options.file,
                                        topics=options.topic,
                                        usecolor=options.color,
                                        deps=options.deps)


class CommandToDoExpirable(CommandToDo):

    def __init__(self, name="todo-expirable", help="List of stale changes that can be expired"):
        super(CommandToDoExpirable, self).__init__(name, help)

    def add_options(self, parser, config):
        super(CommandToDoExpirable, self).add_options(parser, config)

        self.add_option(parser, config,
                        "--age", default="28",
                        help="Set age cutoff in days")

    def get_report(self, config, client, options):
        return ReportToDoListExpirable(client,
                                       age=int(options.age),
                                       projects=self.get_projects(config, options),
                                       branches=options.branch,
                                       files=options.file,
                                       topics=options.topic,
                                       usecolor=options.color,
                                       deps=options.deps)


class CommandComments(CommandCaching):

    def __init__(self, name="comments", help="Display comments on a change"):
        super(CommandComments, self).__init__(name, help)

    def add_options(self, parser, config):
        super(CommandComments, self).add_options(parser, config)

        self.add_option(parser, config,
                        "change", default=None,
                        help="Filter based on status")

        self.add_option(parser, config,
                        "--color", default=False, action="store_true",
                        help="Use terminal color highlighting")

        self.add_option(parser, config,
                        "--all", default=False, action="store_true",
                        help="Don't filter list of comments to strip bots")

        self.add_option(parser, config,
                        "--current", default=False, action="store_true",
                        help="Only display comments against current patch")

        self.add_option(parser, config,
                        "--patch", default=0,
                        help="Only display comments against patch NN")


    @staticmethod
    def wrap_text(message, indent="", width=78):
        lines = textwrap.wrap(message, width)
        return "\n".join(map(lambda x: indent + x, lines))

    @staticmethod
    def format_comments(allcomments, bots, usecolor):
        comments = []
        for comment in allcomments:
            if not comment.is_reviewer_in_list(bots):
                comments.append(comment)

        if len(comments) == 0:
            print (format_color("  No  comments",
                                usecolor,
                                fg="grey"))
            print ("")
            print ("")
        else:
            for comment in comments:
                if comment.file is not None:
                    print ("  %s: (%s) %s:%d" %
                           (format_color(comment.reviewer.name,
                                         usecolor,
                                         styles=["bold"]),
                            comment.reviewer.username,
                            comment.file,
                            comment.line))
                else:
                    print ("  %s: (%s)" %
                           (format_color(comment.reviewer.name,
                                         usecolor,
                                         styles=["bold"]),
                            comment.reviewer.username))
                print ("")
                print (CommandComments.wrap_text(comment.message, "  "))
                print ("")
                print ("")

    @staticmethod
    def format_change(change, bots, usecolor, currentpatch, patchnum):
        print (format_color("Change %s (%s)" % (change.url, change.id),
                            usecolor,
                            fg="red",
                            styles=["bold"]))
        print ("")
        print ("  %s" % change.subject)
        print ("")
        print ("")

        patches = change.patches
        if currentpatch:
            patches = patches[-1:]
        elif patchnum:
            patches = patches[patchnum-1:patchnum]

        for patch in patches:
            print (format_color("Patch %d (%s)" % (patch.number, patch.revision),
                                usecolor,
                                fg="blue",
                                styles=["bold"]))
            print ("")
            comments = []
            comments.extend(patch.comments)

            prefix = "Patch Set %d:" % patch.number
            abandoned = 0
            for comment in change.comments:
                if comment.message.startswith(prefix):
                    comments.append(comment)
                    if comment.message.startswith(prefix + ": Abandoned"):
                        abandoned = patch.number
                elif comment.message.startswith("Restored") and abandoned == patch.number:
                    comments.append(comment)

            CommandComments.format_comments(comments, bots, usecolor)


    def run(self, config, client, options):
        change = options.change

        query = OperationQuery(client,
                               {
                                   "change": [ change ],
                               },
                               patches=OperationQuery.PATCHES_ALL,
                               approvals=True,
                               files=True,
                               comments=True)

        if options.all:
            bots = []
        else:
            bots = config.get_organization_bots()

        def mycb(change):
            self.format_change(change, bots, options.color, options.current, int(options.patch))

        query.run(mycb, limit=1)


class CommandTool(object):

    def __init__(self):
        self.commands = {}

    def add_options(self, parser):
        parser.add_argument("-c", "--config", default=os.path.expanduser("~/.gerrymander"),
                            help=("Override config file (default %s)" %
                                  os.path.expanduser("~/.gerrymander")))
        parser.add_argument("-d", "--debug",
                            help="Display debugging information",
                            action="store_true")
        parser.add_argument("-q", "--quiet",
                            help="Supress display of warnings",
                            action="store_true")

    def add_command(self, subparser, config, cmdclass, name=None, help=None):
        cmd = cmdclass()
        if name is not None:
            cmd.name = name
        if help is not None:
            cmd.help = help

        parser = subparser.add_parser(cmd.name, help=cmd.help)
        cmd.add_options(parser, config)
        parser.set_defaults(func=cmd.execute)
        self.commands[cmd.name] = cmd

    def add_default_commands(self, subparser, config):
        self.add_command(subparser, config, CommandWatch)
        self.add_command(subparser, config, CommandToDoNoones)
        self.add_command(subparser, config, CommandToDoAnyones)
        self.add_command(subparser, config, CommandToDoMine)
        self.add_command(subparser, config, CommandToDoOthers)
        self.add_command(subparser, config, CommandToDoApprovable)
        self.add_command(subparser, config, CommandToDoExpirable)
        self.add_command(subparser, config, CommandPatchReviewStats)
        self.add_command(subparser, config, CommandPatchReviewRate)
        self.add_command(subparser, config, CommandOpenReviewStats)
        self.add_command(subparser, config, CommandChanges)
        self.add_command(subparser, config, CommandComments)

    def add_config_commands(self, subparser, config):
        aliases = config.get_command_aliases()
        for alias in aliases:
            basecmd = config.get_command_alias_basecmd(alias)
            help = config.get_command_alias_help(alias)

            if basecmd not in self.commands:
                raise Exception("Unknown base command '%s'" % basecmd)

            klass = type(self.commands[basecmd])
            self.add_command(subparser, config, klass, alias, help)

    def get_config(self, options):
        return CommandConfig(options.config)

    def execute(self, argv):
        miniparser = argparse.ArgumentParser(add_help=False)
        miniparser.add_argument("-c", "--config", default=os.path.expanduser("~/.gerrymander"),
                                help=("Override config file (default %s)" %
                                      os.path.expanduser("~/.gerrymander")))
        options, remaining = miniparser.parse_known_args(argv)

        config = CommandConfig(options.config)


        parser = argparse.ArgumentParser(description="Gerrymander client")
        self.add_options(parser)
        subparser = parser.add_subparsers()
        subparser.required = True
        subparser.dest = "command"
        self.add_default_commands(subparser, config)
        self.add_config_commands(subparser, config)

        options = parser.parse_args(argv)


        level = logging.WARNING
        if options.debug:
            level = logging.DEBUG
        elif options.quiet:
            level = logging.ERROR

        logging.basicConfig(level=level,
                            format='%(asctime)s %(levelname)s: %(message)s',
                            stream=sys.stderr)

        options.func(config, options)



#
# Copyright (C) 2014 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import os.path
import subprocess
import sys
import json
import hashlib
import time
import fcntl

LOG = logging.getLogger(__name__)

class ClientLive(object):

    def __init__(self, hostname="review", port=None, username=None, keyfile=None):
        self.hostname = hostname
        self.port = port
        self.username = username
        self.keyfile = keyfile

    def _run_async(self, argv):

        def _preexec_fn():
            os.setpgrp()

        stdout = subprocess.PIPE
        stderr = subprocess.PIPE
        LOG.debug("Running cmd %s" % " ".join(argv))
        sp = subprocess.Popen(argv,
                              stdout=stdout,
                              stderr=stderr,
                              stdin=None,
                              preexec_fn=_preexec_fn)
        return sp

    def _build_argv(self, cmdargv):
        argv = ['ssh']
        argv.extend(["-T", "-o", "BatchMode=yes", "-e", "none"])
        if self.port:
            argv.extend(["-p", str(self.port)])
        if self.keyfile and os.path.isfile(self.keyfile):
            argv.extend(['-i', str(self.keyfile)])
        if self.username is not None:
            argv.extend(["%s@%s" % (self.username, self.hostname)])
        else:
            argv.extend([self.hostname])
        argv.extend(["gerrit"])
        argv.extend(cmdargv)
        return argv

    def _process(self, sp, argv, cb):
        while True:
            line = sp.stdout.readline()
            if not line:
                break
            try:
                dec = json.loads(line.decode("UTF-8"))
                if not isinstance(dec, (dict)):
                    raise TypeError("Expected decoded dict, not %s" % (type(dec)))
                cb(dec)
            except Exception:
                LOG.exception("Failure processing %s", line)

        sp.wait()
        if sp.returncode != 0:
            lines = []
            while True:
                line = sp.stderr.readline()
                if not line:
                    break
                lines.append(line.decode("UTF-8"))
            msg = "".join(lines)
            args = " ".join(argv)
            raise Exception("Error running command %s: %s" %
                            (args, msg))

    def run(self, cmdargv, cb):
        argv = self._build_argv(cmdargv)
        sp = self._run_async(argv)
        return self._process(sp, argv, cb)


class ClientCachingLock(object):
    def __init__(self, lockfile):
        self.lockfile = lockfile

    def __enter__(self):
        self.lockfh = open(self.lockfile, "w")
        fcntl.lockf(self.lockfh, fcntl.LOCK_EX)

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            fcntl.lockf(self.lockfh, fcntl.LOCK_UN)
            self.lockfh.close()
            self.lockfh = None
        except:
            LOG.exception("could not release lock on %s" % self.lockfile)


class ClientCaching(ClientLive):

    def __init__(self, hostname="review", port=None, username=None, keyfile=None,
                 cachedir="cache", cachelifetime=86400, refresh=False):
        super(ClientCaching, self).__init__(hostname, port, username, keyfile)
        self.cachedir = cachedir
        self.cachelifetime = cachelifetime
        self.lastpurge = None
        self.refresh = refresh

        if not os.path.exists(self.cachedir):
            os.makedirs(self.cachedir)

    def _purge_cache_locked(self):
        now = time.time()
        if self.lastpurge is not None and (now - self.lastpurge) < 60 * 60:
            return
        self.lastpurge = now

        then = now - self.cachelifetime
        LOG.debug("Looking for files in %s older than %d" % (self.cachedir, then))
        for file in os.listdir(self.cachedir):
            if file == "lock":
                continue
            filepath = os.path.join(self.cachedir, file)
            mtime = os.path.getmtime(filepath)
            LOG.debug("File %s has time %d" % (filepath, mtime))
            if mtime < then:
                LOG.info("Purging outdated cache %s" % filepath)
                os.unlink(filepath)

    def _purge_cache(self):
        # XXX we really need to protect individual files
        # against deletion while they're being read
        lock = ClientCachingLock(os.path.join(self.cachedir, "lock"))
        LOG.debug("acquiring lock for cache")
        with lock as lock:
            self._purge_cache_locked()

    def run(self, cmdargv, cb):
        self._purge_cache()
        argv = self._build_argv(cmdargv)
        args = " ".join(argv)
        m = hashlib.sha256()
        m.update(args.encode("UTF-8"))
        LOG.debug("Finding cache for args '%s'" % args)
        file = self.cachedir + "/" + m.hexdigest() + ".json"
        if not os.path.exists(file) or self.refresh:
            sp = self._run_async(argv)
            try:
                with open(file, "wb") as f:
                    while True:
                        line = sp.stdout.readline()
                        if not line:
                            break
                        f.write(line)
            except:
                os.unlink(file)
                raise

            sp.wait()
            if sp.returncode != 0:
                os.unlink(file)
                lines = []
                while True:
                    line = sp.stderr.readline()
                    if not line:
                        break
                    lines.append(line.decode("UTF-8"))
                msg = "".join(lines)
                raise Exception("Error running command %s: %s" %
                                (args, msg))

        sp = self._run_async(["cat", file])
        self._process(sp, argv, cb)



#
# Copyright (C) 2014 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import collections
import prettytable
import logging
import time
import re
import json
import sys
import xml.dom.minidom

from gerrymander.operations import OperationQuery
from gerrymander.model import ModelApproval
from gerrymander.format import format_date
from gerrymander.format import format_delta
from gerrymander.format import format_title
from gerrymander.format import format_color

LOG = logging.getLogger(__name__)


TreeNode = collections.namedtuple('TreeNode', ['data', 'children'])


class ReportOutputColumn(object):

    ALIGN_LEFT = "l"
    ALIGN_RIGHT = "r"
    ALIGN_CENTER = "c"

    def __init__(self, key, label, mapfunc, sortfunc=None, format=None, truncate=0, align=ALIGN_LEFT, visible=True):
        self.key = key
        self.label = label
        self.mapfunc = mapfunc
        self.sortfunc = sortfunc
        self.format = format
        self.truncate = truncate
        self.align = align
        self.visible = visible

    def get_value(self, report, row):
        val = self.mapfunc(report, self.key, row)
        if self.format is not None:
            val = self.format % val
        elif val is None:
            val = ""

        if type(val) != str:
            try:
                val = val.encode('utf-8')
            except AttributeError:
                val = str(val)

        if self.truncate and len(val) > self.truncate:
            val = val[0:self.truncate] + "..."

        return val

    def get_sort_value(self, report, row):
        if self.sortfunc:
            return self.sortfunc(report, self.key, row)
        else:
            return self.mapfunc(report, self.key, row)

class ReportOutput(object):

    DISPLAY_MODE_TEXT = "text"
    DISPLAY_MODE_CSV = "csv"
    DISPLAY_MODE_XML = "xml"
    DISPLAY_MODE_JSON = "json"

    def __init__(self, usecolor=False):
        super(ReportOutput, self).__init__()
        self.usecolor = usecolor

    def display(self, mode, stream=sys.stdout):
        if mode == ReportOutput.DISPLAY_MODE_TEXT:
            stream.write(self.to_text())
        elif mode == ReportOutput.DISPLAY_MODE_CSV:
            stream.write(self.to_csv())
        elif mode == ReportOutput.DISPLAY_MODE_XML:
            impl = xml.dom.minidom.getDOMImplementation()
            doc = impl.createDocument(None, "report", None)
            self.to_xml(doc, doc.documentElement)
            stream.write(doc.toprettyxml())
        elif mode == ReportOutput.DISPLAY_MODE_JSON:
            doc = []
            self.to_json(doc)
            stream.write(json.dumps(doc, indent=2) + "\n")
        else:
            raise Exception("Unknown display mode '%s'" % mode)

    def to_text(self):
        raise NotImplementedError("Subclass should implement the 'to_text' method")

    def to_csv(self):
        raise NotImplementedError("Subclass should implement the 'to_csv' method")

    def to_xml(self, root):
        raise NotImplementedError("Subclass should implement the 'to_xml' method")

    def to_json(self, root):
        raise NotImplementedError("Subclass should implement the 'to_json' method")


class ReportOutputCompound(ReportOutput):

    def __init__(self):
        self.report = []

    def add_report(self, report):
        self.report.append(report)

    def to_text(self):
        blocks = []
        for report in self.report:
            blocks.append(report.to_text())
        return "\n".join(blocks)

    def to_json(self, root):
        for report in self.report:
            report.to_json(root)

    def to_xml(self, doc, root):
        for report in self.report:
            report.to_xml(doc, root)


class ReportOutputList(ReportOutput):
    def __init__(self, columns, title=None, usecolor=False):
        super(ReportOutputList, self).__init__(usecolor)
        self.columns = columns
        self.row = {}
        self.title = title

    def set_row(self, row):
        self.row = row

    def to_xml(self, doc, root):
        lst = doc.createElement("list")
        root.appendChild(lst)
        if self.title is not None:
            title = doc.createElement("title")
            title.appendChild(doc.createTextNode(self.title))
            lst.appendChild(title)
        headers = doc.createElement("headers")
        content = doc.createElement("content")
        lst.appendChild(headers)
        lst.appendChild(content)

        for col in self.columns:
            if col.visible:
                xmlcol = doc.createElement(col.key)
                xmlcol.appendChild(doc.createTextNode(col.label))
                headers.appendChild(xmlcol)

        for col in self.columns:
            if col.visible:
                xmlfield = doc.createElement(col.key)
                xmlfield.appendChild(doc.createTextNode(col.get_value(self, self.row)))
                content.appendChild(xmlfield)

    def to_json(self, root):
        headers = {}
        for col in self.columns:
            if col.visible:
                headers[col.key] = col.label

        content = {}
        for col in self.columns:
            if col.visible:
                content[col.key] = col.get_value(self, self.row)

        node = {
            "list": {
                "headers": headers,
                "content": content
            }
        }
        if self.title is not None:
            node["list"]["title"] = self.title
        root.append(node)


    def to_text(self):
        labels = []
        width = 1
        for col in self.columns:
            if col.visible:
                if len(col.label) > width:
                    width = len(col.label)
                labels.append(col.label)

        fmt = "  %" + str(width) + "s: %s"
        lines = []
        for col in self.columns:
            if col.visible:
                line = fmt % (col.label, col.get_value(self, self.row))
            lines.append(line)

        prolog = ""
        if self.title is not None:
            prolog = format_title(self.title) + "\n"
        return prolog + "\n".join(lines) + "\n"


class ReportOutputTable(ReportOutput):
    def __init__(self, columns, sortcol, reverse, limit, title=None, usecolor=False):
        super(ReportOutputTable, self).__init__(usecolor)
        self.columns = list(columns)
        self.rows = []
        self.sortcol = sortcol
        self.reverse = reverse
        self.limit = limit
        self.title = title

    def add_column(self, col):
        self.columns.append(col)

    def add_row(self, data, parent=None):
        node = TreeNode(data, [])

        if parent is None:
            self.rows.append(node)
        else:
            parent.children.append(node)
        return node

    def iter_rowlist(self, init_rowlist, output_row, init_data, gen_child_data):
        sortcol = None
        for col in self.columns:
            if col.key == self.sortcol:
                sortcol = col

        total_written = [0]
        def _iter_rowlist(rowlist, data):
            if sortcol is not None:
                rowlist.sort(key = lambda item: sortcol.get_sort_value(
                                 self, item.data),
                             reverse=self.reverse)

            for row in rowlist:
                if self.limit is not None and total_written[0] == self.limit:
                    return
                total_written[0] += 1

                rowdata = output_row(row.data, data)

                if len(row.children) > 0:
                    _iter_rowlist(row.children, gen_child_data(rowdata))
        _iter_rowlist(init_rowlist, init_data)

    def to_xml(self, doc, root):
        table = doc.createElement("table")
        root.appendChild(table)
        if self.title is not None:
            title = doc.createElement("title")
            title.appendChild(doc.createTextNode(self.title))
            table.appendChild(title)
        headers = doc.createElement("headers")
        content = doc.createElement("content")
        table.appendChild(headers)
        table.appendChild(content)

        visible_cols = list(filter(lambda col: col.visible, self.columns))
        for col in visible_cols:
            xmlcol = doc.createElement(col.key)
            xmlcol.appendChild(doc.createTextNode(col.label))
            headers.appendChild(xmlcol)

        def output_xml(row_data, parent):
            xmlrow = doc.createElement("row")
            for col in visible_cols:
                xmlfield = doc.createElement(col.key)
                xmlfield.appendChild(
                    doc.createTextNode(col.get_value(self, row_data)))
                xmlrow.appendChild(xmlfield)
            parent.appendChild(xmlrow)
            return xmlrow

        def gen_children(parent):
            children = doc.createElement("children")
            parent.appendChild(children)
            return children

        self.iter_rowlist(self.rows, output_xml, content, gen_children)

        return doc

    def to_json(self, root):
        visible_cols = list(filter(lambda col: col.visible, self.columns))

        headers = {}
        for col in visible_cols:
            headers[col.key] = col.label

        content = []

        def output_json(row_data, parent):
            data = {}
            for col in visible_cols:
                data[col.key] = col.get_value(self, row_data)
            parent.append(data)
            return data

        def gen_children(parent):
            children = []
            parent['children'] = children
            return children

        self.iter_rowlist(self.rows, output_json, content, gen_children)

        node = {
            "table": {
                "headers": headers,
                "content": content
            }
        }
        if self.title is not None:
            node["table"]["title"] = self.title
        root.append(node)

    def to_text(self):
        labels = []
        visible_cols = list(filter(lambda col: col.visible, self.columns))
        for col in visible_cols:
            labels.append(col.label)
        table = prettytable.PrettyTable(labels)
        for col in visible_cols:
            table.align[col.label] = col.align

        table.padding_width = 1
        indent = 2

        def output_text(row_data, depth):
            fields = list(map(lambda col: col.get_value(self, row_data),
                              visible_cols))
            fields[0] = ' ' * indent * depth + fields[0]
            table.add_row(fields)
            return depth

        def inc_depth(depth):
            return depth + 1

        self.iter_rowlist(self.rows, output_text, 0, inc_depth)

        prolog = ""
        if self.title is not None:
            prolog = format_title(self.title) + "\n"
        return prolog + str(table) + "\n"

    def to_csv(self):
        visible_cols = list(filter(lambda col: col.visible, self.columns))

        lines = []

        if self.title is not None:
            lines.append(self.title)

        labels = []
        for col in visible_cols:
            labels.append(col.label)
        lines.append(",".join(labels))

        # CSV data is flat
        def output_csv(row_data, dummy):
            data = list(map(lambda col: col.get_value(self, row_data), visible_cols))
            lines.append(",".join(data))
            return dummy

        self.iter_rowlist(self.rows, output_csv, None, lambda x: None)

        return "\n".join(lines)


class Report(object):

    def __init__(self, client):
        self.client = client

    def generate(self):
        raise NotImplementedError("Subclass must override generate method")

    def display(self, mode):
        output = self.generate()
        output.display(mode)


class ReportTable(Report):

    def __init__(self, client, columns, sort=None, reverse=False):
        super(ReportTable, self).__init__(client)
        self.columns = columns
        self.limit = None
        self.set_sort_column(sort, reverse)

    def get_columns(self):
        return self.columns

    def get_column(self, key):
        for col in self.columns:
            if col.key == key:
                return col
        return None

    def has_column(self, key):
        col = self.get_column(key)
        if col is None:
            return False
        return True

    def set_sort_column(self, key, reverse=False):
        got = False
        for col in self.columns:
            if col.key == key:
                got = True
        if not got:
            raise Exception("Unknown sort column %s" % key)
        self.sort = key
        self.reverse = reverse

    def set_data_limit(self, limit):
        self.limit = limit

    def new_table(self, title=None):
        return ReportOutputTable(self.columns, self.sort,
                                 self.reverse, self.limit,
                                 title, self.usecolor)


class ReportPatchReviewStats(ReportTable):

    def user_mapfunc(rep, col, row):
        return row[0]

    def team_mapfunc(rep, col, row):
        return row[2]

    def review_mapfunc(rep, col, row):
        return row[1]['total']

    def ratio_mapfunc(rep, col, row):
        plus = float(row[1]['votes']['flag-p2'] + row[1]['votes']['flag-p1'])
        minus = float(row[1]['votes']['flag-m2'] + row[1]['votes']['flag-m1'])
        ratio = (plus / (plus + minus)) * 100
        return ratio

    def vote_mapfunc(rep, col, row):
        return row[1]['votes'][col]

    COLUMNS = [
        ReportOutputColumn("user", "User", user_mapfunc, align=ReportOutputColumn.ALIGN_LEFT),
        ReportOutputColumn("team", "Team", team_mapfunc, align=ReportOutputColumn.ALIGN_LEFT),
        ReportOutputColumn("reviews", "Reviews", review_mapfunc, align=ReportOutputColumn.ALIGN_RIGHT),
        ReportOutputColumn("flag-m2", "-2", vote_mapfunc, align=ReportOutputColumn.ALIGN_RIGHT),
        ReportOutputColumn("flag-m1", "-1", vote_mapfunc, align=ReportOutputColumn.ALIGN_RIGHT),
        ReportOutputColumn("flag-p1", "+1", vote_mapfunc, align=ReportOutputColumn.ALIGN_RIGHT),
        ReportOutputColumn("flag-p2", "+2", vote_mapfunc, align=ReportOutputColumn.ALIGN_RIGHT),
        ReportOutputColumn("ratio", "+/-", ratio_mapfunc, format="%0.0lf%%", align=ReportOutputColumn.ALIGN_RIGHT),
    ]

    def __init__(self, client, projects, maxagedays=30, teams={}, usecolor=False):
        super(ReportPatchReviewStats, self).__init__(client,
                                                     ReportPatchReviewStats.COLUMNS,
                                                     sort="reviews", reverse=True)
        self.projects = projects
        self.teams = teams
        self.maxagedays = maxagedays
        self.usecolor = usecolor

    def generate(self):
        # We could query all projects at once, but if we do them
        # individually it means we get better hit rate against the
        # cache if the report is re-run for many different project
        # combinations
        reviews = []
        cutoff = time.time() - (self.maxagedays * 24 * 60 * 60)
        for project in self.projects:
            query = OperationQuery(self.client,
                                   {
                                       "project": [project],
                                   },
                                   patches=OperationQuery.PATCHES_ALL,
                                   approvals=True)


            def querycb(change):
                for patch in change.patches:
                    for approval in patch.approvals:
                        if approval.is_newer_than(cutoff):
                            reviews.append(approval)

            query.run(querycb)

        reviewers = {}
        for review in reviews:
            if review.action != ModelApproval.ACTION_REVIEWED or review.user is None:
                continue

            reviewer = review.user.username
            if reviewer is None:
                reviewer = review.user.name
                if reviewer is None:
                    continue

            if reviewer.lower() in ["jenkins", "smokestack"]:
                continue

            reviewers.setdefault(reviewer,
                                 {
                                     'votes': {'flag-m2': 0, 'flag-m1': 0, 'flag-p1': 0, 'flag-p2': 0},
                                     'total': 0,
                                 })
            reviewers[reviewer]['total'] = reviewers[reviewer]['total'] + 1
            votes = { "-2" : "flag-m2",
                      "-1" : "flag-m1",
                      "1" : "flag-p1",
                      "2" : "flag-p2" }
            cur = reviewers[reviewer]['votes'][votes[str(review.value)]]
            reviewers[reviewer]['votes'][votes[str(review.value)]] = cur + 1

        compound = ReportOutputCompound()
        table = self.new_table("Review statistics")
        compound.add_report(table)

        for user, votes in reviewers.items():
            userteam = ""
            for team in self.teams.keys():
                if user in self.teams[team]:
                    userteam = team

            table.add_row([user, votes, userteam])

        summary = ReportOutputList([
            ReportOutputColumn("nreviews", "Total reviews", format="%d",
                               mapfunc=lambda rep, col, row: row[0]),
            ReportOutputColumn("nreviewers", "Total rviewers", format="%d",
                               mapfunc=lambda rep, col, row: row[1])
        ], title="Review summary")
        summary.set_row([len(reviews), len(reviewers.keys())])
        compound.add_report(summary)

        return compound


class ReportPatchReviewRate(ReportTable):

    def user_mapfunc(rep, col, row):
        return row[0]

    def team_mapfunc(rep, col, row):
        return row[1]

    def week_mapfunc(rep, col, row):
        if col not in row[2]:
            return 0.0

        return (row[2][col] / 7.0)

    def total_mapfunc(rep, col, row):
        if col not in row[2]:
            return 0.0

        return (row[2][col] / (52.0 * 7.0))

    COLUMNS = [
        ReportOutputColumn("user", "User", user_mapfunc, align=ReportOutputColumn.ALIGN_LEFT),
        ReportOutputColumn("team", "Team", team_mapfunc, align=ReportOutputColumn.ALIGN_LEFT),

        ReportOutputColumn("total", "Total", total_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),

        ReportOutputColumn("week1", "1 week", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week2", "2 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week3", "3 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week4", "4 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),

        ReportOutputColumn("week5", "5 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week6", "6 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week7", "7 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week8", "8 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),

        ReportOutputColumn("week9", "9 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week10", "10 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week11", "11 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week12", "12 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),

        ReportOutputColumn("week13", "13 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week14", "14 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week15", "15 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week16", "16 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),

        ReportOutputColumn("week17", "17 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week18", "18 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week19", "19 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week20", "20 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),

        ReportOutputColumn("week21", "21 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week22", "22 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week23", "23 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week24", "24 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),

        ReportOutputColumn("week25", "25 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week26", "26 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week27", "27 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week28", "28 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),

        ReportOutputColumn("week29", "29 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week30", "30 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week31", "31 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week32", "32 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),

        ReportOutputColumn("week33", "33 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week34", "34 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week35", "35 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week36", "36 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),

        ReportOutputColumn("week37", "37 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week38", "38 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week39", "39 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week40", "40 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),

        ReportOutputColumn("week41", "41 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week42", "42 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week43", "43 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week44", "44 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),

        ReportOutputColumn("week45", "45 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week46", "46 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week47", "47 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week48", "48 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),

        ReportOutputColumn("week49", "49 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week50", "50 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week51", "51 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
        ReportOutputColumn("week52", "52 weeks", week_mapfunc, align=ReportOutputColumn.ALIGN_LEFT, format="%0.2f"),
     ]

    def __init__(self, client, projects, teams={}, usecolor=False):
        super(ReportPatchReviewRate, self).__init__(client,
                                                    ReportPatchReviewRate.COLUMNS,
                                                    sort="total", reverse=True)
        self.projects = projects
        self.teams = teams
        self.usecolor = usecolor

    def generate(self):
        # We could query all projects at once, but if we do them
        # individually it means we get better hit rate against the
        # cache if the report is re-run for many different project
        # combinations
        reviewers = {}
        now = time.time()
        for project in self.projects:
            query = OperationQuery(self.client,
                                   {
                                       "project": [project],
                                   },
                                   patches=OperationQuery.PATCHES_ALL,
                                   approvals=True)


            def querycb(change):
                for patch in change.patches:
                    for approval in patch.approvals:
                        if approval.action == ModelApproval.ACTION_VERIFIED:
                            continue

                        user = approval.user
                        if user is None or user.username is None:
                            continue
                        username = user.username

                        if username not in reviewers:
                            reviewers[username] = { "total": 0}

                        agesecs = approval.get_age(now)
                        ageweeks = int(agesecs / (60 * 60 * 24 * 7)) + 1
                        key = "week%d" % ageweeks

                        if key not in reviewers[username]:
                            reviewers[username][key] = 0

                        reviewers[username][key] = reviewers[username][key] + 1

                        if ageweeks <= 52:
                            reviewers[username]["total"] = reviewers[username]["total"] + 1

            query.run(querycb)

        table = self.new_table("Daily review rates per week")

        for reviewer in reviewers.keys():
            userteam = ""
            for team in self.teams.keys():
                if reviewer in self.teams[team]:
                    userteam = team

            table.add_row([reviewer, userteam, reviewers[reviewer]])

        return table


class ReportBaseChange(ReportTable):

    @staticmethod
    def get_approval_votes(patch):
        # Yes, the numbers are slightly odd order
        # A +2 or -2 more important than any -1 or +1
        # so we prefer them as the summary value
        levels = ["-2", "2", "-1", "1"]

        votes = {
            "c": { "total": collections.defaultdict(int),
                   "list": [],
                   "summary": "",
                   "details": "",
            },
            "v": { "total": collections.defaultdict(int),
                   "list": [],
                   "summary": "",
                   "details": "",
            },
            "w": { "total": collections.defaultdict(int),
                   "list": [],
                   "summary": "",
                   "details": "",
            },
        }

        for approval in patch.approvals:
            got_type = approval.action[0:1].lower()
            if got_type not in votes:
                continue
            vote = str(approval.value)
            votes[got_type]["total"][vote] = votes[got_type]["total"][vote] + 1
            votes[got_type]["list"].append(vote)

        for key in votes.keys():
            votes[key]["details"] = ",".join(votes[key]["list"])

            vals = []
            for level in levels:
                if level in votes[key]["total"]:
                    votes[key]["summary"] = level
                    break

        return votes

    def approvals_mapfunc(rep, col, row):
        patch = row.get_current_patch()
        if patch is None:
            LOG.error("No patch")
            return ""

        votes = ReportBaseChange.get_approval_votes(patch)
        keys = list(votes.keys())
        keys.sort(reverse=True)

        data = " ".join(map(lambda val: "%s=%s" % (val, votes[val]["details"]), keys))

        if rep.usecolor:
            if votes["w"]["total"]["1"] > 0: # Stuff pending merge
                return format_color(data, fg="blue", styles=["bold"])
            elif votes["w"]["total"]["-1"] > 0: # Work-in-progress
                return format_color(data, fg="magenta", styles=[])
            elif votes["c"]["total"]["-2"] > 0: # Hard-nack from core
                return format_color(data, fg="red", styles=["bold"])
            elif votes["c"]["total"]["-1"] > 0 or votes["v"]["total"]["-1"] > 0: # Nack from any or bots
                return format_color(data, fg="red", styles=[])
            elif votes["c"]["total"]["2"] > 0: # Approval from core
                return format_color(data, fg="green", styles=["bold"])
            elif votes["c"]["total"]["1"] > 0: # Approval from any
                return format_color(data, fg="green", styles=[])
            else:
                return data
        else:
            return data

    def votes_mapfunc(rep, col, row):
        patch = row.get_current_patch()
        if patch is None:
            LOG.error("No patch")
            return ""

        if col == "tests":
            coltype = "v"
        elif col == "reviews":
            coltype = "c"
        else:
            coltype = "w"

        votes = ReportBaseChange.get_approval_votes(patch)
        data = "%2s" % votes[coltype]["summary"]
        if rep.usecolor:
            if votes[coltype]["total"]["-2"] > 0: # Hard-nack from core
                return format_color(data, fg="red", styles=["bold"])
            elif votes[coltype]["total"]["2"] > 0: # Approval from core
                return format_color(data, fg="green", styles=["bold"])
            elif votes[coltype]["total"]["-1"] > 0: # Soft-nack from any
                return format_color(data, fg="red", styles=[])
            elif votes[coltype]["total"]["1"] > 0: # Approval from any
                return format_color(data, fg="green", styles=[])
            else:
                return data
        else:
            return data

    def user_mapfunc(rep, col, row):
        if not row.owner or not row.owner.username:
            return "<unknown>"
        return row.owner.username

    def date_mapfunc(rep, col, row):
        if col == "lastUpdated":
            return format_date(row.lastUpdated)
        else:
            return format_date(row.createdOn)

    def date_sortfunc(rep, col, row):
        if col == "lastUpdated":
            return row.lastUpdated
        else:
            return row.createdOn

    COLUMNS = [
        ReportOutputColumn("status", "Status", lambda rep, col, row: row.status),
        ReportOutputColumn("topic", "Topic", lambda rep, col, row: row.topic, visible=False),
        ReportOutputColumn("url", "URL", lambda rep, col, row: row.url),
        ReportOutputColumn("owner", "Owner", user_mapfunc),
        ReportOutputColumn("project", "Project", lambda rep, col, row: row.project, visible=False),
        ReportOutputColumn("branch", "Branch", lambda rep, col, row: row.branch, visible=False),
        ReportOutputColumn("subject", "Subject", lambda rep, col, row: row.subject, truncate=30),
        ReportOutputColumn("createdOn", "Created", date_mapfunc, date_sortfunc),
        ReportOutputColumn("lastUpdated", "Updated", date_mapfunc, date_sortfunc),
        ReportOutputColumn("approvals", "Approvals", approvals_mapfunc, visible=False),
        ReportOutputColumn("tests", "Tests", votes_mapfunc),
        ReportOutputColumn("reviews", "Reviews", votes_mapfunc),
        ReportOutputColumn("workflow", "Workflow", votes_mapfunc),
    ]

    def __init__(self, client, usecolor=False):
        super(ReportBaseChange, self).__init__(client, ReportBaseChange.COLUMNS,
                                               sort="createdOn", reverse=False)
        self.usecolor = usecolor


class ReportChangeList(ReportBaseChange):

    def __init__(self, client, usecolor, title,
                 query_terms, patches, files=None, rawquery=None,
                 deps=False):
        super(ReportChangeList, self).__init__(client, usecolor)
        self.title = title
        self.query_terms = query_terms
        self.patches = patches
        self.rawquery = rawquery
        self.files = files
        self.deps = deps

    def filter(self, change):
        return True

    def generate(self):
        query = OperationQuery(self.client,
                               self.query_terms,
                               rawquery=self.rawquery,
                               patches=self.patches,
                               approvals=True,
                               files=(self.files is not None),
                               deps=self.deps)

        def match_files(change):
            if len(self.files) == 0:
                return True
            for filere in self.files:
                for file in change.get_current_patch().files:
                    if re.search(filere, file.path):
                        return True
            return False

        table = self.new_table(self.title)

        if not self.deps:
            def querycb(change):
                if self.filter(change) and match_files(change):
                    table.add_row(change)
            query.run(querycb)
        else:
            # Index all changes by change id
            nodes = {}
            def querycb(change):
                if self.filter(change) and match_files(change):
                    nodes[change.id] = TreeNode(change, [])
            query.run(querycb)

            # Create a hierarchy of changes by dependency
            root = []
            for node in nodes.values():
                change = node.data
                parent = nodes.get(change.depends)

                # Check if we've seen this change's parent
                if parent is None:
                    # If not, add it to the root list
                    root.append(node)
                else:
                    # Add it as a child
                    parent.children.append(node)

            # Add changes recursively to output table
            def add_nodes(change_nodes, parent=None):
                for change_node in change_nodes:
                    row_node = table.add_row(change_node.data, parent)
                    add_nodes(change_node.children, row_node)
            add_nodes(root)

        return table


class ReportChanges(ReportChangeList):

    def __init__(self, client, projects=[], owners=[],
                 status=[], messages=[], branches=[], topics=[], reviewers=[],
                 approvals=[], files=[], rawquery=None, usecolor=False,
                 deps=False):
        self.approvals = approvals

        query_terms = {
            "project": projects,
            "owner": owners,
            "message": messages,
            "branch": branches,
            "topic": topics,
            "status": status,
            "reviewer": reviewers,
        }
        super(ReportChanges, self).__init__(client, usecolor, "Changes",
                                            query_terms,
                                            OperationQuery.PATCHES_CURRENT,
                                            files=files,
                                            rawquery=rawquery, deps=deps)


class ReportToDoList(ReportChangeList):

    def __init__(self, client, projects=[], branches=[],
                 files=[], topics=[], reviewers=[], usecolor=False, deps=False):
        query_terms = {
            "project": projects,
            "status": [ OperationQuery.STATUS_OPEN ],
            "branch": branches,
            "topic": topics,
            "reviewer": reviewers,
        }
        super(ReportToDoList, self).__init__(client, usecolor,
                                             "Changes To Do List",
                                             query_terms,
                                             OperationQuery.PATCHES_ALL,
                                             files, deps=deps)


class ReportToDoListMine(ReportToDoList):

    def __init__(self, client, username, projects=[],
                 branches=[], files=[], topics=[], usecolor=False, deps=False):
        '''
        Report to provide a list of changes 'username' has
        reviewed an older version of the patch, and needs
        to provide feedback on latest version
        '''
        super(ReportToDoListMine, self).__init__(client,
                                                 projects,
                                                 reviewers=[ username ],
                                                 branches=branches,
                                                 files=files,
                                                 topics=topics,
                                                 usecolor=usecolor,
                                                 deps=deps)
        self.username = username

    def filter(self, change):
        if change.has_owner([self.username]):
            return False
        if not change.has_current_reviewers([self.username]):
            return True
        return False


class ReportToDoListOthers(ReportToDoList):
    def __init__(self, client, username, bots=[], projects=[],
                 branches=[], files=[], topics=[], usecolor=False, deps=False):
        '''
        Report to provide a list of changes where 'username' has
        never reviewed, but at least one other non-bot user has
        provided review
        '''
        super(ReportToDoListOthers, self).__init__(client,
                                                   projects,
                                                   reviewers=[ "!", username ],
                                                   branches=branches,
                                                   files=files,
                                                   topics=topics,
                                                   usecolor=usecolor,
                                                   deps=deps)
        self.bots = bots

    def filter(self, change):
        # allchanges contains changes where 'username' has
        # not reviewed any version of the patch. We want to
        # filter out changes which only have bots, or have
        # no reviewers at all.
        if change.has_owner([self.username]):
            return False
        if change.has_any_other_reviewers(self.bots):
            return True
        return False


class ReportToDoListAnyones(ReportToDoList):

    def __init__(self, client, username, bots=[], projects=[],
                 branches=[], files=[], topics=[], usecolor=False, deps=False):
        '''
        Report to provide a list of changes where at least
        one other non-bot user has provided review
        '''
        super(ReportToDoListAnyones, self).__init__(client,
                                                    projects,
                                                    branches=branches,
                                                    files=files,
                                                    topics=topics,
                                                    usecolor=usecolor,
                                                    deps=deps)
        self.bots = bots
        self.username = username

    def filter(self, change):
        if change.has_owner([self.username]):
            return False
        if change.has_current_reviewers([self.username]):
            return False
        if change.has_any_other_reviewers(self.bots):
            return True
        return False


class ReportToDoListNoones(ReportToDoList):

    def __init__(self, client, bots=[], projects=[],
                 branches=[], files=[], topics=[], usecolor=False, deps=False):
        '''
        Report to provide a list of changes that no one
        has ever reviewed
        '''
        super(ReportToDoListNoones, self).__init__(client,
                                                   projects,
                                                   branches=branches,
                                                   files=files,
                                                   topics=topics,
                                                   usecolor=usecolor,
                                                   deps=deps)
        self.bots = bots

    def filter(self, change):
        if change.has_owner([self.username]):
            return False
        if not change.has_any_other_reviewers(self.bots):
            return True
        return False


class ReportToDoListApprovable(ReportToDoList):

    def __init__(self, client, username, strict, projects=[],
                 branches=[], files=[], topics=[], usecolor=False, deps=False):
        '''
        Report to provide a list of changes that no one
        has ever reviewed
        '''
        super(ReportToDoListApprovable, self).__init__(client,
                                                       projects,
                                                       branches=branches,
                                                       files=files,
                                                       topics=topics,
                                                       usecolor=usecolor,
                                                       deps=deps)
        self.username = username
        self.strict = strict

    def filter(self, change):
        if change.has_owner([self.username]):
            return False
        if (change.has_current_approval(ModelApproval.ACTION_REVIEWED, 2) and
            not change.has_current_approval(ModelApproval.ACTION_WORKFLOW, -1) and
            not change.has_current_approval(ModelApproval.ACTION_WORKFLOW, 1) and
            not change.has_current_approval(ModelApproval.ACTION_REVIEWED, -2) and
            not change.has_current_reviewers([self.username])):

            if (self.strict and
                change.has_current_approval(ModelApproval.ACTION_REVIEWED, -1)):
                return False

            return True
        return False


class ReportToDoListExpirable(ReportToDoList):

    def __init__(self, client, age=28, projects=[],
                 branches=[], files=[], topics=[], usecolor=False, deps=False):
        '''
        Report to provide a list of changes that are
        stale and can potentially be expired
        '''
        super(ReportToDoListExpirable, self).__init__(client,
                                                      projects,
                                                      branches=branches,
                                                      files=files,
                                                      topics=topics,
                                                      usecolor=usecolor,
                                                      deps=deps)
        self.age = age

    def filter(self, change):
        if change.get_current_reviewer_nack_age() > (self.age * 24 * 60 * 60):
            return True
        return False


class ReportOpenReviewStats(ReportBaseChange):

    def __init__(self, client, projects, branch="master", topic="", days=7, usecolor=False):
        super(ReportOpenReviewStats, self).__init__(client, usecolor)
        self.projects = projects
        self.branch = branch
        self.topic = topic
        self.days = days

    @staticmethod
    def average_age(changes, ages):
        if len(changes) == 0:
            return 0
        total = 0
        for change in changes:
            total += ages[change]
        return format_delta(total / len(changes))

    @staticmethod
    def median_age(changes, ages):
        if len(changes) == 0:
            return 0
        total = 0
        wantages = []
        for change in changes:
            wantages.append(ages[change])
        wantages.sort()
        return format_delta(wantages[int(len(wantages)/2)])

    @staticmethod
    def older_than(changes, ages, cutoffdays):
        cutoff = cutoffdays * 24 * 60 * 60
        older = 0
        for change in changes:
            if ages[change] > cutoff:
                older = older + 1
        return older

    @staticmethod
    def get_longest_changes(ids, changes, ages, count):
        want = []
        for id in sorted(ids, key=lambda x: ages[x]):
            want.append(changes[id])
        return want

    def generate(self):
        # We could query all projects at once, but if we do them
        # individually it means we get better hit rate against the
        # cache if the report is re-run for many different project
        # combinations
        agecurrent = {}
        agefirst = {}
        agenonnacked = {}
        wait_reviewer = []
        wait_submitter = []
        changes = {}
        for project in self.projects:
            query = OperationQuery(self.client,
                                   {
                                       "project": [project],
                                       "status": [OperationQuery.STATUS_OPEN],
                                       "branch": [self.branch],
                                       "topic": [self.topic],
                                   },
                                   patches=OperationQuery.PATCHES_ALL,
                                   approvals=True)


            def querycb(change):
                if change.status != "NEW":
                    return

                now = time.time()
                current = change.get_current_patch()
                first = change.get_first_patch()
                nonnacked = change.get_reviewer_not_nacked_patch()

                changes[change.id] = change

                if current.is_nacked():
                    wait_submitter.append(change.id)
                else:
                    wait_reviewer.append(change.id)

                agecurrent[change.id] = current.get_age(now)
                agefirst[change.id] = first.get_age(now)
                if nonnacked:
                    agenonnacked[change.id] = nonnacked.get_age(now)
                else:
                    agenonnacked[change.id] = 0

            query.run(querycb)

        compound = ReportOutputCompound()
        summary = ReportOutputList([
            ReportOutputColumn("nreviews", "Total open reviews", format="%d",
                               mapfunc=lambda rep, col, row: row[0] + row [1]),
            ReportOutputColumn("waitsubmitter", "Waiting on submitter", format="%d",
                               mapfunc=lambda rep, col, row: row[0]),
            ReportOutputColumn("waitreviewer", "Waiting on reviewer", format="%d",
                               mapfunc=lambda rep, col, row: row[1]),
        ], title="Review summary")
        summary.set_row([len(wait_submitter), len(wait_reviewer)])
        compound.add_report(summary)

        lastrev = ReportOutputList([
            ReportOutputColumn("average", "Average wait time",
                               mapfunc=lambda rep, col, row: row[0]),
            ReportOutputColumn("median", "Median wait time",
                               mapfunc=lambda rep, col, row: row[1]),
            ReportOutputColumn("stale", "Older than %d days" % self.days, format="%d",
                               mapfunc=lambda rep, col, row: row[2]),
        ], title="Summary since current revision")
        lastrev.set_row([self.average_age(wait_reviewer, agecurrent),
                         self.median_age(wait_reviewer, agecurrent),
                         self.older_than(wait_reviewer, agecurrent, self.days)])
        compound.add_report(lastrev)


        firstrev = ReportOutputList([
            ReportOutputColumn("average", "Average wait time",
                               mapfunc=lambda rep, col, row: row[0]),
            ReportOutputColumn("median", "Median wait time",
                               mapfunc=lambda rep, col, row: row[1]),
        ], title="Summary since first revision")
        firstrev.set_row([self.average_age(wait_reviewer, agefirst),
                          self.median_age(wait_reviewer, agefirst)])
        compound.add_report(firstrev)


        nonnackedrev = ReportOutputList([
            ReportOutputColumn("average", "Average wait time",
                               mapfunc=lambda rep, col, row: row[0]),
            ReportOutputColumn("median", "Median wait time",
                               mapfunc=lambda rep, col, row: row[1]),
        ], title="Summary since last revision without -1/-2 from reviewer")
        nonnackedrev.set_row([self.average_age(wait_reviewer, agenonnacked),
                              self.median_age(wait_reviewer, agenonnacked)])
        compound.add_report(nonnackedrev)


        def waitlastmap(rep, col, row):
            return format_delta(row.get_current_age())

        def waitlastsort(rep, col, row):
            return row.get_current_age()

        waitlastrev = self.new_table("Longest waiting since current revision")
        waitlastrev.add_column(ReportOutputColumn("age", "Age",
                                                  sortfunc=waitlastsort,
                                                  mapfunc=waitlastmap))
        waitlastrev.sortcol = "age"
        waitlastrev.reverse = True
        for change in self.get_longest_changes(wait_reviewer, changes, agecurrent, 5):
            waitlastrev.add_row(change)
        compound.add_report(waitlastrev)


        def waitfirstmap(rep, col, row):
            return format_delta(row.get_first_age())

        def waitfirstsort(rep, col, row):
            return row.get_first_age()

        waitfirstrev = self.new_table("Longest waiting since first revision")
        waitfirstrev.add_column(ReportOutputColumn("age", "Age",
                                                   sortfunc=waitfirstsort,
                                                   mapfunc=waitfirstmap))
        waitfirstrev.sortcol = "age"
        waitfirstrev.reverse = True
        for change in self.get_longest_changes(wait_reviewer, changes, agefirst, 5):
            waitfirstrev.add_row(change)
        compound.add_report(waitfirstrev)


        def waitnonnackedmap(rep, col, row):
            return format_delta(row.get_reviewer_not_nacked_age())

        def waitnonnackedsort(rep, col, row):
            return row.get_reviewer_not_nacked_age()

        waitnonnackedrev = self.new_table("Longest waiting since last revision without -1/-2 from reviewer")
        waitnonnackedrev.add_column(ReportOutputColumn("age", "Age",
                                                       sortfunc=waitnonnackedsort,
                                                       mapfunc=waitnonnackedmap))
        waitnonnackedrev.sortcol = "age"
        waitnonnackedrev.reverse = True
        for change in self.get_longest_changes(wait_reviewer, changes, agenonnacked, 5):
            waitnonnackedrev.add_row(change)
        compound.add_report(waitnonnackedrev)

        return compound



#
# Copyright (C) 2014 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import subprocess
try:
    # py3
    from urllib.parse import urlparse
except:
    # py2
    from urlparse import urlparse

def get_git_config(key):
    '''Read a git configuration value use "git config --get ..."'''
    try:
        val = subprocess.check_output([
            'git', 'config', '--get', key])
    except subprocess.CalledProcessError:
        return None

    if type(val) == bytes:
        return val.decode(encoding="utf-8")
    else:
        return val

def get_remote_info(remote):
    '''Read information for the named remote from the git configuration
    and return a (user, host, port) tuple.'''
    url = get_git_config('remote.%s.url' % remote)
    if not url:
        return (None, None, None)

    # only ssh urls make sense.  arguably this should support the
    # user@host:path syntax as well, but remotes configured using
    # "git review -s" will never look like that.
    if not url.startswith('ssh://'):
        return (None, None, None)

    url = urlparse(url)

    try:
        userhost, port = url.netloc.split(':')
    except ValueError:
        port = None
        userhost = url.netloc

    try:
        user, host = userhost.split('@')
    except ValueError:
        user = None
        host = url.netloc

    return (user, host, port)



#
# Copyright (C) 2014 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from gerrymander.model import ModelChange
from gerrymander.model import ModelEvent


class OperationBase(object):

    def __init__(self, client):
        self.client = client


class OperationQuery(OperationBase):
    PATCHES_NONE = "none"
    PATCHES_CURRENT = "current"
    PATCHES_ALL = "all"

    STATUS_SUBMITTED = "submitted"
    STATUS_REVIEWED = "reviewed"
    STATUS_MERGED = "merged"
    STATUS_ABANDONED = "abandoned"
    STATUS_OPEN = "open"
    STATUS_CLOSED = "closed"

    def __init__(self, client, terms={}, rawquery=None, patches=PATCHES_NONE,
                 approvals=False, files=False, comments=False, deps=False):
        OperationBase.__init__(self, client)
        self.terms = terms
        self.rawquery = rawquery
        self.patches = patches
        self.approvals = approvals
        self.files = files
        self.comments = comments
        self.deps = deps

        if self.patches == OperationQuery.PATCHES_NONE:
            if self.approvals:
                raise Exception("approvals cannot be requested without patches")
            if self.files:
                raise Exception("files cannot be requested without patches")

    def get_args(self, limit=None, offset=None, sortkey=None):
        args = ["query", "--format=JSON"]
        if self.patches == OperationQuery.PATCHES_CURRENT:
            args.append("--current-patch-set")
        elif self.patches == OperationQuery.PATCHES_ALL:
            args.append("--patch-sets")

        if self.approvals:
            args.append("--all-approvals")
        if self.files:
            args.append("--files")
        if self.comments:
            args.append("--comments")
        if self.deps:
            args.append("--dependencies")

        clauses = []
        if offset is not None:
            args.append("--start")
            args.append("%d" % offset)

        if limit is not None:
            clauses.append("limit:" + str(limit))
        if sortkey is not None:
            clauses.append("resume_sortkey:" + sortkey)
        if self.rawquery is not None:
            clauses.append("(" + self.rawquery + ")")
        terms = list(self.terms.keys())
        terms.sort()
        for term in terms:
            negateAll = False
            terms = self.terms[term]
            if len(terms) > 0 and terms[0] == "!":
                negateAll = True
                terms = terms[1:]

            if len(terms) == 0:
                continue

            subclauses = []
            for value in terms:
                subclauses.append("%s:%s" % (term, value))

            clause = " OR ".join(subclauses)
            if negateAll:
                clause = "( NOT ( " + clause + " ) )"
            else:
                clause = "( " + clause + " )"
            clauses.append(clause)
        args.append(" AND ".join(clauses))
        return args

    def run(self, cb, limit=None):
        class tracker(object):
            def __init__(self):
                self.gotany = True
                self.count = 0
                self.sortkey = None
                self.has_more = False

        c = tracker()
        def mycb(line):
            if 'rowCount' in line:
                # New gerrit sets 'moreChanges'
                if 'moreChanges' in line:
                    c.has_more = line['moreChanges']
                return

            if 'type' in line and line['type'] == "error":
                raise Exception(line['message'])

            change = ModelChange.from_json(line)
            # Old gerrit sets 'sortKey'
            if "sortKey" in line:
                c.sortkey = line["sortKey"]
            c.gotany = True
            c.count = c.count + 1
            cb(change)

        if limit is None:
            while c.gotany:
                c.gotany = False
                offset = None
                if c.has_more:
                    offset = c.count
                self.client.run(self.get_args(500, offset, c.sortkey), mycb)
                if not c.sortkey and not c.has_more:
                    break
        else:
            while c.count < limit and c.gotany:
                want = limit - c.count
                if want > 500:
                    want = 500
                c.gotany = False
                offset = None
                if c.has_more:
                    offset = c.count
                self.client.run(self.get_args(want, offset, c.sortkey), mycb)
                if not c.sortkey and not c.has_more:
                    break
        return 0


class OperationWatch(OperationBase):

    def __init__(self, client):
        OperationBase.__init__(self, client)

    def run(self, cb):
        def mycb(line):
            event = ModelEvent.from_json(line)
            if event:
                cb(event)

        return self.client.run(["stream-events"], mycb)



#
# Copyright (C) 2014 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import unittest

from gerrymander.model import ModelChange, ModelUser, ModelPatch, ModelApproval

JSON_CHANGE='''
{
    "project": "openstack/nova-specs",
    "branch": "master",
    "topic": "bp/is",
    "id": "Idcf7b35c93422d0d13c3ffb33aca3c5eca4f3c94",
    "number": "85556",
    "subject": "Libvirt-Enable suppport for discard option for disk device",
    "owner": {
        "name": "boh.ricky",
        "email": "boh.ricky@gmail.com",
        "username": "boh.ricky"
    },
    "url": "https://review.openstack.org/85556",
    "createdOn": 1396715237,
    "lastUpdated": 1397487650,
    "sortKey": "002c6ba400014e34",
    "open": true,
    "status": "NEW",
    "patchSets": [
        {
            "number": "1",
            "revision": "b2bc7a9f5feab51b61fc9c7725251868e998703d",
            "ref": "refs/changes/56/85556/1",
            "uploader": {
                "name": "boh.ricky",
                "email": "boh.ricky@gmail.com",
                "username": "boh.ricky"
            },
            "createdOn": 1396715237,
            "approvals": [
                {
                    "type": "Verified",
                    "description": "Verified",
                    "value": "-1",
                    "grantedOn": 1396715281,
                    "by": {
                        "name": "Jenkins",
                        "username": "jenkins"
                    }
                }
            ]
        },
        {
            "number": "2",
            "revision": "1ea93fd986afabc479bf93a4ce0c6df4924a7c9e",
            "ref": "refs/changes/56/85556/2",
            "uploader": {
                "name": "boh.ricky",
                "email": "boh.ricky@gmail.com",
                "username": "boh.ricky"
            },
            "createdOn": 1396716807,
            "approvals": [
                {
                    "type": "Verified",
                    "description": "Verified",
                    "value": "1",
                    "grantedOn": 1397487649,
                    "by": {
                        "name": "Jenkins",
                        "username": "jenkins"
                    }
                },
                {
                    "type": "CRVW",
                    "description": "Code Review",
                    "value": "-1",
                    "grantedOn": 1397040232,
                    "by": {
                        "name": "Daniel Berrange",
                        "email": "berrange@redhat.com",
                        "username": "berrange"
                    }
                },
                {
                    "type": "CRVW",
                    "description": "Code Review",
                    "value": "-1",
                    "grantedOn": 1397137466,
                    "by": {
                        "name": "Dan Smith",
                        "email": "dms@danplanet.com",
                        "username": "danms"
                    }
                }
            ]
        }
    ]
}'''

class TestGerrymanderModel(unittest.TestCase):

    def test_json_parse(self):
        change = ModelChange.from_json(json.loads(JSON_CHANGE))

        self.assertEqual(type(change), ModelChange)

        self.assertEqual(change.project, "openstack/nova-specs")
        self.assertEqual(change.branch, "master")
        self.assertEqual(change.number, 85556)
        self.assertEqual(change.subject, "Libvirt-Enable suppport for discard option for disk device")
        self.assertEqual(change.url, "https://review.openstack.org/85556")
        self.assertEqual(change.createdOn, 1396715237)
        self.assertEqual(change.lastUpdated, 1397487650)
        self.assertEqual(change.status, "NEW")

        self.assertEqual(type(change.owner), ModelUser)
        self.assertEqual(change.owner.name, "boh.ricky")
        self.assertEqual(change.owner.username, "boh.ricky")
        self.assertEqual(change.owner.email, "boh.ricky@gmail.com")

        self.assertEqual(len(change.patches), 2)
        self.assertEqual(change.patches[0].number, 1)
        self.assertEqual(change.patches[0].revision, "b2bc7a9f5feab51b61fc9c7725251868e998703d")
        self.assertEqual(change.patches[0].ref, "refs/changes/56/85556/1")
        self.assertEqual(change.patches[0].createdOn, 1396715237)

        self.assertEqual(type(change.patches[0].uploader), ModelUser)
        self.assertEqual(change.patches[0].uploader.name, "boh.ricky")
        self.assertEqual(change.patches[0].uploader.username, "boh.ricky")
        self.assertEqual(change.patches[0].uploader.email, "boh.ricky@gmail.com")

        self.assertEqual(len(change.patches[0].approvals), 1)
        self.assertEqual(type(change.patches[0].approvals[0]), ModelApproval)
        self.assertEqual(change.patches[0].approvals[0].action, ModelApproval.ACTION_VERIFIED)
        self.assertEqual(change.patches[0].approvals[0].description, "Verified")
        self.assertEqual(change.patches[0].approvals[0].value, -1)
        self.assertEqual(change.patches[0].approvals[0].grantedOn, 1396715281)

        self.assertEqual(type(change.patches[0].approvals[0].user), ModelUser)
        self.assertEqual(change.patches[0].approvals[0].user.name, "Jenkins")
        self.assertEqual(change.patches[0].approvals[0].user.username, "jenkins")



#
# Copyright (C) 2014 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time

class ModelBase(object):
    pass

class ModelUser(ModelBase):
    def __init__(self, name, email=None, username=None):
        self.name = name
        self.email = email
        self.username = username

    def is_in_list(self, users):
        if self.name is not None and self.name in users:
            return True
        if self.username is not None and self.username in users:
            return True
        return False

    @staticmethod
    def from_json(data):
        return ModelUser(data.get("name", None),
                         data.get("email", None),
                         data.get("username", None))


class ModelFile(ModelBase):
    ACTION_MODIFIED = "MODIFIED"
    ACTION_ADDED = "ADDED"
    ACTION_DELETED = "DELETED"
    ACTION_RENAMED = "RENAMED"

    def __init__(self, path, action):
        self.path = path
        self.action = action

    @staticmethod
    def from_json(data):
        return ModelFile(data.get("file", None),
                         data.get("type", None))


class ModelApproval(ModelBase):
    ACTION_VERIFIED = "Verified"
    ACTION_REVIEWED = "Code-Review"
    ACTION_WORKFLOW = "Workflow"

    def __init__(self, action, value, description, grantedOn=None, user=None):
        self.action = action
        self.value = value
        self.description = description
        if grantedOn is not None:
            self.grantedOn = int(grantedOn)
        else:
            self.grantedOn = None
        self.user = user

    def is_user_in_list(self, users):
        if self.user is None:
            return False
        return self.user.is_in_list(users)

    def is_newer_than(self, then):
        if self.grantedOn is None:
            return False
        if self.grantedOn > then:
            return True
        return False

    def is_nack(self):
        if self.value < 0:
            return True
        return False

    def is_reviewer_nack(self):
        if self.action != ModelApproval.ACTION_REVIEWED:
            return False
        if self.value < 0:
            return True
        return False

    def get_age(self, now=None):
        if now is None:
            now = time.time()
        return now - self.grantedOn

    @staticmethod
    def from_json(data):
        user = None
        if data.get("by", None):
            user = ModelUser.from_json(data["by"])
        return ModelApproval(data.get("type", None),
                             int(data.get("value", 0)),
                             data.get("description", None),
                             data.get("grantedOn", None),
                             user)

class ModelComment(ModelBase):
    def __init__(self, message, file, line, reviewer):
        self.message = message
        self.file = file
        self.line = line
        self.reviewer = reviewer

    def is_reviewer_in_list(self, users):
        if self.reviewer is None:
            return False
        return self.reviewer.is_in_list(users)

    @staticmethod
    def from_json(data):
        user = None
        if data.get("reviewer", None):
            user = ModelUser.from_json(data["reviewer"])
        return ModelComment(data.get("message", ""),
                            data.get("file", None),
                            data.get("line", 0),
                            user)


class ModelPatch(ModelBase):

    def __init__(self, number, revision, ref, uploader, createdOn, approvals=[], files=[], comments=[]):
        self.number = number
        self.revision = revision
        self.ref = ref
        self.uploader = uploader
        self.createdOn = createdOn
        self.approvals = approvals
        self.files = files
        self.comments = comments

    @staticmethod
    def is_user_in_list(users, user):
        if user.username is not None and user.username in users:
            return True

        if user.email is not None and user.email in users:
            return True

        return False

    def is_reviewer_nacked(self):
        for approval in self.approvals:
            if approval.is_reviewer_nack():
                return True
        return False

    def is_nacked(self):
        for approval in self.approvals:
            if approval.is_nack():
                return True
        return False

    def get_reviewer_nack_age(self, now):
        age = 0
        for approval in self.approvals:
            if approval.is_reviewer_nack():
                thisage = approval.get_age(now)
                if (age == 0) or (thisage < age):
                    age = thisage
        return age

    def get_age(self, now):
        if len(self.approvals) == 0:
            return now - self.createdOn

        age = 0
        for approval in self.approvals:
            thisage = now - approval.grantedOn
            if thisage > age:
                age = thisage
        return age

    def has_other_reviewers(self, excludeusers):
        '''Determine if the patch has been reviewed by any
        users that are not in 'excludeusers'''

        hasReviewers = False
        for approval in self.approvals:
            if not approval.is_user_in_list(excludeusers):
                hasReviewers = True
        return hasReviewers

    def has_reviewers(self, includeusers):
        '''Determine if the patch has been reviewed by any
        users that are in 'includeusers'''

        hasReviewers = False
        for approval in self.approvals:
            if approval.user is None:
                continue
            if approval.is_user_in_list(includeusers):
                hasReviewers = True
        return hasReviewers

    def has_current_approval(self, action, value):
        '''Determine if the change has an approval vote
        of the requested type and value'''
        for approval in self.approvals:
            if (approval.action == action and
                approval.value == value):
                return True
        return False

    @staticmethod
    def from_json(data):
        files = []
        for f in data.get("files", []):
            files.append(ModelFile.from_json(f))

        approvals = []
        for f in data.get("approvals", []):
            approvals.append(ModelApproval.from_json(f))

        user = None
        if "uploader" in data:
            user = ModelUser.from_json(data["uploader"])

        comments = []
        for c in data.get("comments", []):
            comments.append(ModelComment.from_json(c))

        return ModelPatch(int(data.get("number", 0)),
                          data.get("revision"),
                          data.get("ref"),
                          user,
                          data.get("createdOn"),
                          approvals,
                          files,
                          comments)


class ModelChange(ModelBase):

    def __init__(self, project, branch, topic, id, number, subject, owner, url,
                 createdOn, lastUpdated, status,
                 patches = [], comments = [], depends = None):
        self.project = project
        self.branch = branch
        self.topic = topic
        self.id = id
        self.number = number
        self.subject = subject
        self.owner = owner
        self.url = url
        if createdOn is not None:
            self.createdOn = int(createdOn)
        else:
            self.createdOn = None
        if lastUpdated is not None:
            self.lastUpdated = int(lastUpdated)
        else:
            self.lastUpdated = None
        self.status = status
        self.patches = patches
        self.comments = comments
        self.depends = depends

    def get_current_patch(self):
        if len(self.patches) == 0:
            return None
        return self.patches[len(self.patches) - 1]

    def get_first_patch(self):
        if len(self.patches) == 0:
            return None
        return self.patches[0]

    def get_reviewer_not_nacked_patch(self):
        prev = None
        for patch in reversed(self.patches):
            if patch.is_reviewer_nacked():
                break
            prev = patch
        return prev

    def get_current_age(self):
        patch = self.get_current_patch()
        return patch.get_age(time.time())

    def get_first_age(self):
        patch = self.get_first_patch()
        return patch.get_age(time.time())

    def get_reviewer_not_nacked_age(self):
        patch = self.get_reviewer_not_nacked_patch()
        if patch is None:
            return 0
        return patch.get_age(time.time())

    def get_current_reviewer_nack_age(self):
        patch = self.get_current_patch()
        if patch is None:
            return 0
        return patch.get_reviewer_nack_age(time.time())

    @staticmethod
    def is_user_in_list(users, user):
        if user.username is not None and user.username in users:
            return True

        if user.email is not None and user.email in users:
            return True

        return False

    def has_any_other_reviewers(self, excludeusers):
        '''Determine if any patch in the change has been
        reviewed by any user not in the list of 'excludeusers'''

        hasReviewers = False
        for patch in self.patches:
            if patch.has_other_reviewers(excludeusers):
                hasReviewers = True
        return hasReviewers

    def has_any_reviewers(self, includeusers):
        '''Determine if any patch in the change has been
        reviewed by any user in the list of 'includeusers'''

        hasReviewers = False
        for patch in self.patches:
            if patch.has_reviewers(includeusers):
                hasReviewers = True
        return hasReviewers

    def has_current_reviewers(self, includeusers):
        '''Determine if the current patch version has
        been reviewed by any of the users in 'includeusers'. '''
        patch = self.get_current_patch()
        if patch is None:
            return False
        return patch.has_reviewers(includeusers)

    def has_current_other_reviewers(self, excludeusers):
        '''Determine if the current patch version has
        been reviewed by any of the users not in 'excludeusers'. '''
        patch = self.get_current_patch()
        if patch is None:
            return False
        return patch.has_other_reviewers(excludeusers)

    def has_owner(self, includeusers):
        '''Determine if the change is owned by anyone
        in 'incldueusers' list.'''
        return self.is_user_in_list(includeusers, self.owner)

    def has_current_approval(self, action, value):
        '''Determine if the change has an approval vote
        of the requested type and value'''
        patch = self.get_current_patch()
        if patch is None:
            return False
        return patch.has_current_approval(action, value)

    @staticmethod
    def from_json(data):
        patches = []
        for p in data.get("patchSets", []):
            patches.append(ModelPatch.from_json(p))

        user = None
        if "owner" in data:
            user = ModelUser.from_json(data["owner"])

        number = None
        if "number" in data:
            number = int(data.get("number"))

        comments = []
        for c in data.get("comments", []):
            comments.append(ModelComment.from_json(c))

        depends = None
        dependsChange = data.get("dependsOn")
        if dependsChange is not None and len(dependsChange) > 0:
            # We only track a single dependency
            depends = dependsChange[0].get("id")

        return ModelChange(data.get("project", None),
                           data.get("branch", None),
                           data.get("topic", None),
                           data.get("id", None),
                           number,
                           data.get("subject", None),
                           user,
                           data.get("url", None),
                           data.get("createdOn", None),
                           data.get("lastUpdated", None),
                           data.get("status", None),
                           patches,
                           comments,
                           depends)


class ModelEvent(ModelBase):

    def __init__(self, change, patch, user):
        self.change = change
        self.patch = patch
        self.user = user

    def is_user_in_list(self, users):
        if self.user is None:
            return False
        return self.user.is_in_list(users)

    @staticmethod
    def from_json(data):
        if data["type"] == "comment-added":
            return ModelEventCommentAdd.from_json(data)
        elif data["type"] == "patchset-created":
            return ModelEventPatchCreate.from_json(data)
        elif data["type"] == "change-merged":
            return ModelEventChangeMerge.from_json(data)
        elif data["type"] == "change-abandoned":
            return ModelEventChangeAbandon.from_json(data)
        elif data["type"] == "change-restored":
            return ModelEventChangeRestore.from_json(data)
        elif data["type"] == "ref-updated":
            return ModelEventRefUpdated.from_json(data)
        elif data["type"] == "reviewer-added":
            return ModelEventReviewerAdded.from_json(data)
        elif data["type"] == "topic-changed":
            return ModelEventTopicChanged.from_json(data)
        else:
            raise Exception("Unknown event '%s'" % data["type"])


class ModelEventCommentAdd(ModelEvent):

    def __init__(self, change, patch, user, comment, approvals):
        ModelEvent.__init__(self, change, patch, user)
        self.comment = comment
        self.approvals = approvals

    @staticmethod
    def from_json(data):
        change = ModelChange.from_json(data["change"])
        patch = ModelPatch.from_json(data["patchSet"])
        user = ModelUser.from_json(data["author"])
        comment = data["comment"]
        approvals = []
        for f in data.get("approvals", []):
            approvals.append(ModelApproval.from_json(f))
        return ModelEventCommentAdd(change, patch, user, comment, approvals)


class ModelEventPatchCreate(ModelEvent):

    def __init__(self, change, patch, user):
        ModelEvent.__init__(self, change, patch, user)

    @staticmethod
    def from_json(data):
        change = ModelChange.from_json(data["change"])
        patch = ModelPatch.from_json(data["patchSet"])
        user = ModelUser.from_json(data["uploader"])
        return ModelEventPatchCreate(change, patch, user)


class ModelEventChangeMerge(ModelEvent):

    def __init__(self, change, patch, user):
        ModelEvent.__init__(self, change, patch, user)

    @staticmethod
    def from_json(data):
        change = ModelChange.from_json(data["change"])
        patch = ModelPatch.from_json(data["patchSet"])
        user = ModelUser.from_json(data["submitter"])
        return ModelEventChangeMerge(change, patch, user)


class ModelEventChangeAbandon(ModelEvent):

    def __init__(self, change, patch, user):
        ModelEvent.__init__(self, change, patch, user)

    @staticmethod
    def from_json(data):
        change = ModelChange.from_json(data["change"])
        user = ModelUser.from_json(data["abandoner"])
        return ModelEventChangeAbandon(change, None, user)


class ModelEventChangeRestore(ModelEvent):

    def __init__(self, change, patch, user):
        ModelEvent.__init__(self, change, patch, user)

    @staticmethod
    def from_json(data):
        change = ModelChange.from_json(data["change"])
        user = ModelUser.from_json(data["restorer"])
        return ModelEventChangeRestore(change, None, user)


class ModelEventReviewerAdded(ModelEvent):

    def __init__(self, change, patch, user):
        ModelEvent.__init__(self, change, patch, user)

    @staticmethod
    def from_json(data):
        change = ModelChange.from_json(data["change"])
        user = ModelUser.from_json(data["reviewer"])
        return ModelEventReviewerAdded(change, None, user)


class ModelEventTopicChanged(ModelEvent):

    def __init__(self, change, patch, user):
        ModelEvent.__init__(self, change, patch, user)

    @staticmethod
    def from_json(data):
        change = ModelChange.from_json(data["change"])
        user = ModelUser.from_json(data["changer"])
        return ModelEventTopicChanged(change, None, user)


class ModelEventRefUpdated(ModelEvent):

    def __init__(self, change, patch, user):
        ModelEvent.__init__(self, change, patch, user)

    @staticmethod
    def from_json(data):
        submitter = data.get("submitter", None)
        user = None
        if submitter is not None:
            user = ModelUser.from_json(submitter)
        return ModelEventRefUpdated(None, None, user)




from setuptools import setup
from setuptools import Command
from distutils.command.build import build
from distutils.util import get_platform

import sys
import os
import re
import time

class my_build(build):
    user_options = build.user_options

    description = "build everything needed to install"

    def gen_rpm_spec(self):
        f1 = open('gerrymander.spec.in', 'r')
        f2 = open('gerrymander.spec', 'w')
        for line in f1:
            f2.write(line
                     .replace('@PY_VERSION@', self.distribution.get_version()))
        f1.close()
        f2.close()

    def gen_authors(self):
        f = os.popen("git log --pretty=format:'%aN <%aE>'")
        authors = []
        for line in f:
            authors.append("   " + line.strip())

        authors.sort(key=str.lower)

        f1 = open('AUTHORS.in', 'r')
        f2 = open('AUTHORS', 'w')
        for line in f1:
            f2.write(line.replace('@AUTHORS@', "\n".join(set(authors))))
        f1.close()
        f2.close()


    def gen_changelog(self):
        f1 = os.popen("git log '--pretty=format:%H:%ct %an  <%ae>%n%n%s%n%b%n'")
        f2 = open("ChangeLog", 'w')

        for line in f1:
            m = re.match(r'([a-f0-9]+):(\d+)\s(.*)', line)
            if m:
                t = time.gmtime(int(m.group(2)))
                f2.write("%04d-%02d-%02d %s\n" % (t.tm_year, t.tm_mon, t.tm_mday, m.group(3)))
            else:
                if re.match(r'Signed-off-by', line):
                    continue
                f2.write("    " + line.strip() + "\n")

        f1.close()
        f2.close()


    def run(self):
        if not os.path.exists("build"):
            os.mkdir("build")

        if os.path.exists(".git"):
            try:
                self.gen_rpm_spec()
                self.gen_authors()
                self.gen_changelog()

                build.run(self)

            except:
                files = ["gerrymander.spec",
                         "AUTHORS",
                         "ChangeLog"]
                for f in files:
                    if os.path.exists(f):
                        os.unlink(f)
        else:
            build.run(self)


class my_test(Command):
    user_options = [
        ('build-base=', 'b',
         "base directory for build library"),
        ('build-platlib=', None,
         "build directory for platform-specific distributions"),
        ('plat-name=', 'p',
         "platform name to build for, if supported "
         "(default: %s)" % get_platform()),
    ]

    description = "Run test suite."

    def initialize_options(self):
        self.build_base = 'build'
        self.build_platlib = None
        self.plat_name = None

    def finalize_options(self):
        if self.plat_name is None:
            self.plat_name = get_platform()

        plat_specifier = ".%s-%s" % (self.plat_name, sys.version[0:3])

        if hasattr(sys, 'gettotalrefcount'):
            plat_specifier += '-pydebug'

        if self.build_platlib is None:
            self.build_platlib = os.path.join(self.build_base,
                                              'lib' + plat_specifier)

    def run(self):
        """
        Run test suite
        """

        if "PYTHONPATH" in os.environ:
            os.environ["PYTHONPATH"] = self.build_platlib + ":" + os.environ["PYTHONPATH"]
        else:
            os.environ["PYTHONPATH"] = self.build_platlib
        self.spawn([sys.executable, "/usr/bin/nosetests"])


setup(
    name="gerrymander",
    version="1.5",
    description="A client API and command line tool for gerrit",
    long_description="Gerrymander provides a set of APIs and command "
    "line tool for interacting with the gerrit code review system",
    author="Daniel P. Berrange",
    author_email="dan-gerrymander@berrange.com",
    license="ASL 2.0",
    url="https://github.com/berrange/gerrymander",
    scripts=([
        "scripts/gerrymander",
        ]),
    packages=["gerrymander"],
    install_requires=["prettytable>=0.5"],
    cmdclass = {
          'build': my_build,
          'test': my_test
    },
)
